"""
Comparison splinebox and scipy: contour approximation
-----------------------------------------------------

This example compares splinebox and scipy when trying to approximate a contour/shape
with a closed spline with a fixe number of control points.
"""

import matplotlib.pyplot as plt
import numpy as np
import scipy
import skimage
import splinebox

# %%
# Let's load the astronaut example image from skimage and crop out a single coins using slicing.
img = skimage.data.coins()[90:150, 240:300]
plt.imshow(img, cmap="gray")
plt.show()

# %%
# Next, we will segment the coin using Otsu's method and fix some holes in the mask.
thresh = skimage.filters.threshold_otsu(img)
mask = img > thresh
mask = skimage.morphology.remove_small_holes(mask)
plt.imshow(mask, cmap="gray")
plt.show()

# %%
# We can find the contour of the coin using find contours.
# The return value is a list of contours. Since our binary image only
# has one contour we select the first element of the list.
contours = skimage.measure.find_contours(mask)
contour = contours[0]
plt.imshow(mask, cmap="gray", alpha=0.5)
plt.plot(contour[:, 1], contour[:, 0])
plt.show()

# %%
# For closed contours, the last point is identical to the first one, so we will
# chop it off.
contour = contour[:-1]

# %%
# Our goal is to fit a cubic B-spline with M control points to
# the contour. We will first use splinbox to acchive this and then
# use scipy.
M = 9

spline = splinebox.Spline(M=M, basis_function=splinebox.B3(), closed=True)
spline.fit(contour)

ts = np.linspace(0, M, 100)
splinebox_vals = spline.eval(ts)
splinebox_control_points = spline.control_points

plt.imshow(mask, cmap="gray", alpha=0.5)
# plt.scatter(contours[0][:, 1], contours[0][:, 0], label="contour")
plt.plot(splinebox_vals[:, 1], splinebox_vals[:, 0], label="splinebox")
plt.scatter(splinebox_control_points[:, 1], splinebox_control_points[:, 0])
plt.legend()
plt.show()

# %%
# Let see how this compares to scipy...

# Number of data points
N = len(contour)
k = 3

# %%
# To get a spline with a specific number of control points in scipy
# we have to precalculate the parameters values `t` for the knots and the parameter values `u`
# for the data points. It is important that we account for the periodicity and padding of the knots.
t = np.arange(-k, M + k + 1)
u = np.linspace(0, M, N + 1)[:-1]

# %%
# When constructinc the spline using `splprep` we have to specify the oder of the basis spline,
# the `u` and `t` we just computed, and the periodicity. Since we don't want to smooth our fit
# but instead regularize it by fixing the number of control points we need to set `s=0` and
# `task=-1`.
tck, u = scipy.interpolate.splprep(contour.T, k=k, u=u, t=t, task=-1, s=0, per=N)
ts = np.linspace(0, M, 100)
scipy_vals = scipy.interpolate.splev(ts, tck)
scipy_control_points = tck[1]

plt.imshow(mask, cmap="gray", alpha=0.5)
# plt.scatter(contours[0][:, 1], contours[0][:, 0], label="contour")
plt.plot(scipy_vals[1], scipy_vals[0], label="scipy")
plt.scatter(scipy_control_points[1], scipy_control_points[0])
plt.legend()
plt.show()



"""
Generate the splinebox logo
---------------------------

This example shows how the `S` of the splinebox logo is
constructed using a B3 spline.
"""

import matplotlib.pyplot as plt
import numpy as np
import splinebox.basis_functions
import splinebox.spline_curves

points = np.array([[1.5, 1.5], [1, 2], [0, 2], [0, 1], [1, 1], [1, 0], [0, 0], [-0.5, 0.5]])
points = points[:, ::-1]
M = len(points)
padded_points = np.concatenate(
    [
        np.array([[1.5, 1.5]]),
        points,
        np.array([[0.5, -0.5]]),
    ]
)
basis_function = splinebox.basis_functions.B3()
spline = splinebox.spline_curves.Spline(M=M, basis_function=basis_function, closed=False, control_points=padded_points)

t = np.linspace(0.8, M - 1.5, 100)

vals = spline.eval(t)
curvature = spline.curvature(t)
normals = spline.normal(t)

# You can limit the max height of the curvature comb using the factor d.
max_comb_height = 0.3
d = max_comb_height / np.max(np.abs(curvature))

comb = vals + d * curvature[:, np.newaxis] * normals

# color1 = "#f5bc47"
color1 = "#568b22"
color2 = "#568b22"
linewidth1 = 3
linewidth2 = 6

plt.plot(comb[:, 1], comb[:, 0], label="Curvature comb", color=color1, linewidth=linewidth1)
for p in range(len(comb)):
    plt.plot([vals[p, 1], comb[p, 1]], [vals[p, 0], comb[p, 0]], color=color1, linewidth=linewidth1)
plt.plot(vals[:, 1], vals[:, 0], linewidth=linewidth2, color=color2)
plt.gca().set_aspect("equal", "box")
plt.savefig("box.pdf")
plt.show()



"""
Comparison splinebox and scipy: edge fitting
--------------------------------------------

This example compares splinebox and scipy when trying to fit a spline
to an end of an image.
"""

# sphinx_gallery_thumbnail_number = 2

import cycler
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import scipy
import scipy.optimize
import skimage
import splinebox

# %%
# First, we will set some defaults to style our plot.
# You can ingnore this section if you don't care about the style
# of the plots.
mpl.rcParams["image.cmap"] = "gray"
mpl.rcParams["lines.linewidth"] = 2
mpl.rcParams["axes.prop_cycle"] = cycler.cycler(color=("r",))

# %%
# In this example, we will use a crop of the scikit-image's `text` example image.
img = skimage.data.text()
img = img[45:78, 300:430]
plt.imshow(img)
plt.show()

# %%
# Our goal is to fit a spline to the integral sign on the page.
# We won't cover how to automatically find a good initial guess but instead
# just select five pixels that are more or less equally spaced along the
# integral as the initial not of our spline.
initial_knots = np.array([[24, 12], [23, 47], [16, 72], [8, 97], [8, 124]])

# %%
# In order to evaluate how well our spline fits the line, we will have to
# interpolate the pixel values at non-integer position.
# Since the line is black on a white background, our goal is to move the
# spline in a way that minimises the average pixel value under it.
# This is commonly refered to as the image energy.
interpolator = scipy.interpolate.RectBivariateSpline(np.arange(img.shape[0]), np.arange(img.shape[1]), img, s=1)

# %%
# We will start by constructing an initial spline using
# splinebox.
M = len(initial_knots)
basis_function = splinebox.B3()
spline = splinebox.Spline(M, basis_function, closed=False)
spline.knots = initial_knots

# %%
# Let's define the parameter values at which we want to sample the spline.
# Here, we chose to sample 50 points inbetween two knots.
t = np.linspace(0, M - 1, M * 50)

# %%
# In order to compar the fitted spline to the intial one,
# we save it's positions and knots for plotting later on.
initial_vals = spline.eval(t)
initial_knots = spline.knots


# %%
# Before we can start fitting the spline, we have to define the loss function.
# In this example our loss function consists of two parts, the image energy,
# discussed above and an internal energy. Here, we choose to use the curvilinear
# reparametrization energy as our internal energy. It promotes equidistant
# spacing of the knots in terms of arc length. In practice, this avoids sharp
# bends and stops the spline from looping/folding back on itself.
# Without it, the image energy would reward the spline for visiting the darkest pixels
# multiple times.
# The parameter alpha can be used balance the contribution of the image and internal energies.
def loss_function_splinebox(control_points, alpha):
    spline.control_points = control_points.reshape((-1, 2))
    coordinates = spline.eval(t)
    image_energy = np.mean(interpolator(coordinates[:, 0], coordinates[:, 1], grid=False))
    internal_energy = spline.curvilinear_reparametrization_energy()
    return image_energy + alpha * internal_energy


# %%
# For the fitting procedure we can simply use scipy.
# The value for alpha was empirically set to 500.
initial_control_points = spline.control_points
scipy.optimize.minimize(loss_function_splinebox, initial_control_points.flatten(), args=(500,))

# %%
# Let's plot the results.
fitted_vals = spline.eval(t)
fitted_knots = spline.knots

fix, axes = plt.subplots(2, 1, sharex=True, sharey=True)
axes[0].imshow(img)
axes[0].plot(initial_vals[:, 1], initial_vals[:, 0], label="initial spline")
axes[0].scatter(initial_knots[:, 1], initial_knots[:, 0], label="initial knots")
axes[0].legend()
axes[1].imshow(img)
axes[1].plot(fitted_vals[:, 1], fitted_vals[:, 0], label="fitted spline")
axes[1].scatter(fitted_knots[:, 1], fitted_knots[:, 0], label="fitted knots")
axes[1].legend()
plt.suptitle("SplineBox")
plt.tight_layout()
plt.show()

# %%
# Next, we will try to acchive the same fit using scipy.
# As before, we begin by construcing an initial spline using the initial knots.
# NOTE: you have to set `bc_type` in order to get a spline with the desired
# number of knots.
k = 3
t_knots = np.arange(M)
spline = scipy.interpolate.make_interp_spline(t_knots, initial_knots, k=3, bc_type="natural")

initial_vals = spline(t)
initial_knots = spline(spline.t)[k:-k]


# %%
# Since scipy does not provide a function to compute the curvilinear reparametrization
# energy, we have to do it ourselfs.
def loss_function_scipy(control_points, alpha):
    spline.c = control_points.reshape((-1, 2))
    coordinates = spline(t)
    image_energy = np.mean(interpolator(coordinates[:, 0], coordinates[:, 1], grid=False))
    derivative = spline.derivative()
    integral = scipy.integrate.quad(lambda t: np.linalg.norm(derivative(t)), 0, M - 1)
    length = integral[0]
    c = (length / M) ** 2
    integral = scipy.integrate.quad(lambda t: (np.linalg.norm(derivative(t)) ** 2 - c) ** 2, 0, M - 1)
    internal_energy = integral[0] / length**4
    return image_energy + alpha * internal_energy


initial_control_points = spline.c
scipy.optimize.minimize(loss_function_scipy, initial_control_points.flatten(), args=(500,))

# %%
# Let's take a look at the results.
fitted_vals = spline(t)
fitted_knots = spline(spline.t[k:-k])

fix, axes = plt.subplots(2, 1, sharex=True, sharey=True)
axes[0].imshow(img, cmap="gray")
axes[0].plot(initial_vals[:, 1], initial_vals[:, 0], label="initial spline")
axes[0].scatter(initial_knots[:, 1], initial_knots[:, 0], label="initial knots")
axes[0].legend()
axes[1].imshow(img, cmap="gray")
axes[1].plot(fitted_vals[:, 1], fitted_vals[:, 0], label="fitted spline")
axes[1].scatter(fitted_knots[:, 1], fitted_knots[:, 0], label="fitted knots")
axes[1].legend()
plt.suptitle("SciPy")
plt.tight_layout()
plt.show()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

import cycler

project = "splinebox"
copyright = "2024, Florian Aymanns, Virginie Uhlmann, Edward Ando"  # noqa: A001
author = "Florian Aymanns, Virginie Uhlmann, Edward Ando"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.mathjax",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "matplotlib.sphinxext.plot_directive",
    "sphinx_gallery.gen_gallery",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"
html_title = "Splinebox Documentation"
html_logo = "_static/logo.png"
html_favicon = "_static/favicon.png"
html_sourcelink_suffix = ""
html_short_title = "SplineBox"

html_theme_options = {
    "header_links_before_dropdown": 6,
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/EPFL-Center-for-Imaging/splinebox",
            "icon": "fa-brands fa-github",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/splinebox/",
            "icon": "fa-brands fa-python",
        },
        {
            "name": "Contact",
            "url": "mailto: florian.aymanns@epfl.ch",
            "icon": "fa-brands fa-telegram",
        },
        {
            "name": "EPFL Center for Imaging",
            "url": "https://imaging.epfl.ch/",
            "icon": "_static/imaging.png",
            "type": "local",
        },
    ],
    "use_edit_page_button": True,
    "show_toc_level": 2,
    "navbar_align": "content",  # [left, content, right] For testing that the navbar items align properly
    "navbar_center": ["navbar-nav"],
    "navbar_start": ["navbar-logo", "version-switcher"],
    "navbar_end": ["theme-switcher", "navbar-icon-links"],
    "secondary_sidebar_items": ["page-toc", "searchbox", "edit-this-page", "sourcelink"],
    "footer_start": ["copyright"],
    "pygment_light_style": "tango",
}

html_context = {
    "github_user": "EPFL-Center-for-Imaging",
    "github_repo": "splinebox",
    "github_version": "main",
    "doc_path": "doc",
    "default_mode": "auto",
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["css/custom.css"]

# Example gallery configuration
sphinx_gallery_conf = {
    "examples_dirs": "../examples",  # path to your example scripts
    "gallery_dirs": "auto_examples",  # path to where to save gallery generated output
}

# Matplotlib sphinxext configuration
plot_html_show_formats = False
plot_html_show_source_link = False
plot_formats = [
    ("png", 500),
]
plot_rcparams = {"axes.prop_cycle": cycler.cycler(color=["#228b18"])}



"""
Closed interpolating splines
============================

This example shows different closed interpolating splines.
"""

import matplotlib.pyplot as plt
import numpy as np
import splinebox.basis_functions
import splinebox.spline_curves

n, m = 2, 2
fig, axes = plt.subplots(n, m, sharex=True, sharey=True)

# Some pixel coordinates to interpolate
coordinates = np.array([[0, 100], [50, 100], [100, 50], [100, 0], [100, -100], [50, -50], [0, -100], [-100, 0]])

for i in range(n):
    for j in range(m):
        axes[i, j].scatter(coordinates[:, 0], coordinates[:, 1])

# The parameter values at which the spline is evaluated
ts = np.linspace(0, len(coordinates), 1000)

for i, (name, basis_function) in enumerate(
    (
        ("B1", splinebox.basis_functions.B1()),
        ("B3", splinebox.basis_functions.B3()),
        ("Exponential", splinebox.basis_functions.Exponential(len(coordinates))),
        ("CatmullRom", splinebox.basis_functions.CatmullRom()),
    )
):
    curve = splinebox.spline_curves.Spline(len(coordinates), basis_function, True)
    curve.knots = coordinates
    discreteContour = curve.eval(ts)

    axes[i // n, i % n].plot(discreteContour[:, 0], discreteContour[:, 1])
    axes[i // n, i % n].set_title(name)
    axes[i // n, i % n].set_aspect("equal", adjustable="box")

plt.show()



import itertools
import math
import unittest.mock

import numpy as np
import pytest
import splinebox.basis_functions
import splinebox.spline_curves


@pytest.fixture(
    params=[
        splinebox.basis_functions.B1,
        splinebox.basis_functions.B2,
        splinebox.basis_functions.B3,
        splinebox.basis_functions.Exponential,
        splinebox.basis_functions.CatmullRom,
        splinebox.basis_functions.CubicHermite,
        splinebox.basis_functions.ExponentialHermite,
    ]
)
def basis_function(request, M):
    basis_function = request.param
    params = {}
    if basis_function in [splinebox.basis_functions.Exponential, splinebox.basis_functions.ExponentialHermite]:
        params["M"] = M
    return basis_function(**params)


@pytest.fixture(
    params=[
        (splinebox.basis_functions.B1, {}),
        (splinebox.basis_functions.B2, {}),
        (splinebox.basis_functions.B3, {}),
        (splinebox.basis_functions.Exponential, {"M": 5}),
        (splinebox.basis_functions.CatmullRom, {}),
    ],
    ids=["B1", "B2", "B3", "Exponential-M5", "CatmullRom"],
)
def non_hermite_basis_function(request):
    basis_function, params = request.param
    return basis_function(**params)


@pytest.fixture(
    params=[
        (splinebox.basis_functions.CubicHermite, {}),
        (splinebox.basis_functions.ExponentialHermite, {"M": 5}),
    ],
    ids=["CubicHermite", "ExponentialHermite"],
)
def hermite_basis_function(request):
    basis_function, params = request.param
    return basis_function(**params)


@pytest.fixture
def is_locally_refinable():
    def _is_locally_refinable(basis_function):
        return not isinstance(
            basis_function,
            (
                splinebox.basis_functions.CubicHermite,
                splinebox.basis_functions.ExponentialHermite,
                splinebox.basis_functions.CatmullRom,
            ),
        )

    return _is_locally_refinable


@pytest.fixture(params=[4, 5, 8])
def M(request):
    return request.param


@pytest.fixture(params=[True, False])
def closed(request):
    return request.param


@pytest.fixture(params=[None, splinebox.spline_curves.padding_function])
def padding_function(request):
    return request.param


@pytest.fixture
def spline_curve(basis_function, M, closed, padding_function):
    if isinstance(
        basis_function, (splinebox.basis_functions.CubicHermite, splinebox.basis_functions.ExponentialHermite)
    ):
        spline = splinebox.spline_curves.HermiteSpline(
            M, basis_function, closed=closed, padding_function=padding_function
        )
        spline._check_control_points_and_tangents = unittest.mock.MagicMock()
        spline._check_control_points = unittest.mock.MagicMock()
    else:
        spline = splinebox.spline_curves.Spline(M, basis_function, closed=closed, padding_function=padding_function)
        spline._check_control_points = unittest.mock.MagicMock()
    return spline


@pytest.fixture
def non_hermite_spline_curve(non_hermite_basis_function, M, closed):
    return splinebox.spline_curves.Spline(M, non_hermite_basis_function, closed=closed)


@pytest.fixture
def hermite_spline_curve(hermite_basis_function, M, closed):
    return splinebox.spline_curves.HermiteSpline(M, hermite_basis_function, closed=closed)


@pytest.fixture
def closed_spline_curve(basis_function, M):
    if hasattr(basis_function, "M"):
        M = basis_function.M
    if isinstance(
        basis_function, (splinebox.basis_functions.CubicHermite, splinebox.basis_functions.ExponentialHermite)
    ):
        return splinebox.spline_curves.HermiteSpline(M, basis_function, closed=True)
    return splinebox.spline_curves.Spline(M, basis_function, closed=True)


@pytest.fixture
def open_spline_curve(basis_function, M):
    if hasattr(basis_function, "M"):
        M = basis_function.M
    if isinstance(
        basis_function, (splinebox.basis_functions.CubicHermite, splinebox.basis_functions.ExponentialHermite)
    ):
        return splinebox.spline_curves.HermiteSpline(M, basis_function, closed=False)
    return splinebox.spline_curves.Spline(M, basis_function, closed=False)


@pytest.fixture(params=[0, 1, 2])
def derivative(request):
    return request.param


@pytest.fixture(params=[1.5, np.linspace(0, 10, 1000)])
def eval_positions(request):
    return request.param


@pytest.fixture(params=[1, 2, 3, 5])
def codomain_dimensionality(request):
    """
    The number of dimensions the spline output
    has. If the dimension is 2, a curve spline
    describes a line in 2D space parameterized by
    a single parameter.
    """
    return request.param


@pytest.fixture
def coeff_gen(codomain_dimensionality):
    rng = np.random.default_rng(seed=1492)

    def _coeff_gen(M, support, closed):
        if closed:
            points = rng.random((M, codomain_dimensionality)) * 100
        else:
            points = rng.random((M + 2 * (math.ceil(support / 2) - 1), codomain_dimensionality)) * 100
        # remove superfluos dimension if codomain_dimensionlity is 1
        return np.squeeze(points)

    return _coeff_gen


@pytest.fixture
def knot_gen(codomain_dimensionality):
    rng = np.random.default_rng(seed=2657)

    def _knot_gen(M=100):
        return np.squeeze(rng.random((M, codomain_dimensionality)))

    return _knot_gen


@pytest.fixture
def is_spline():
    def _is_spline(obj):
        return isinstance(obj, (splinebox.spline_curves.Spline, splinebox.spline_curves.HermiteSpline))

    return _is_spline


@pytest.fixture
def is_interpolating(is_spline):
    def _is_interpolating(obj):
        basis_function = obj.basis_function if is_spline(obj) else obj
        return np.allclose(basis_function.eval(0), 1)

    return _is_interpolating


@pytest.fixture
def not_differentiable_twice(is_spline):
    def _not_differentiable_twice(obj):
        basis_function = obj.basis_function if is_spline(obj) else obj
        return isinstance(
            basis_function,
            (
                splinebox.basis_functions.B1,
                splinebox.basis_functions.CubicHermite,
                splinebox.basis_functions.ExponentialHermite,
            ),
        )

    return _not_differentiable_twice


@pytest.fixture
def is_hermite_spline():
    def _is_hermite_spline(spline_curve):
        return isinstance(
            spline_curve,
            splinebox.spline_curves.HermiteSpline,
        )

    return _is_hermite_spline


@pytest.fixture
def translation_vector(codomain_dimensionality):
    return np.random.rand(codomain_dimensionality) * 10 - 5


@pytest.fixture
def initialized_spline_curve(spline_curve, is_hermite_spline, coeff_gen):
    support = spline_curve.basis_function.support
    closed = spline_curve.closed
    M = spline_curve.M

    spline_curve.control_points = coeff_gen(M, support, closed)
    if is_hermite_spline(spline_curve):
        spline_curve.tangents = coeff_gen(spline_curve.M, support, closed)

    return spline_curve


@pytest.fixture
def rotation_matrix(codomain_dimensionality):
    rng = np.random.default_rng(seed=9577)

    R = np.eye(codomain_dimensionality)
    if codomain_dimensionality == 1:
        return R
    # Get the axis of all possible rotation planes aligned with the axis
    plane_indices = list(itertools.combinations(np.arange(codomain_dimensionality), 2))
    # randomly choose an angle between -pi and pi for each rotation plane
    thetas = rng.random(len(plane_indices)) * 2 * np.pi - np.pi
    for (i, j), theta in zip(plane_indices, thetas):
        R_theta = np.eye(codomain_dimensionality)
        R_theta[i, i] = np.cos(theta)
        R_theta[i, j] = np.sin(theta)
        R_theta[j, i] = -np.sin(theta)
        R_theta[j, j] = np.cos(theta)
        R = R @ R_theta
    return R


@pytest.fixture(params=[False, True])
def arc_length_parametrization(request):
    return request.param


@pytest.fixture(params=[2, 20, 50])
def n_points(request):
    return request.param


@pytest.fixture
def points(codomain_dimensionality, n_points):
    rng = np.random.default_rng(seed=5544)
    return np.squeeze(rng.random((n_points, codomain_dimensionality)))



import matplotlib.pyplot as plt
import numpy as np
import splinebox.basis_functions

fig, axes = plt.subplots(3, 1, sharex=True)

b2 = splinebox.basis_functions.B2()

t = np.linspace(-2, 2, 100)

b2_0th = b2.eval(t)
b2_1st = b2.eval(t, derivative=1)
b2_2nd = b2.eval(t, derivative=2)


fig.suptitle("B2 basis function and its derivatives")
axes[0].plot(t, b2_0th, label=r"$\Phi(t)$")
axes[0].legend()
axes[1].plot(t, b2_1st, label=r"$\frac{d\Phi}{dt}(t)$")
axes[1].legend()
axes[2].plot(t, b2_2nd, label=r"$\frac{d^2\Phi}{dt^2}(t)$")
axes[2].legend()
axes[2].set_xlabel(r"$t$")
plt.tight_layout()

plt.show()



import matplotlib.pyplot as plt
import numpy as np
import splinebox.basis_functions

fig, axes = plt.subplots(3, 1, sharex=True)

M = 5
basis = splinebox.basis_functions.Exponential(M=M)

t = np.linspace(-3, 3, 100)

basis_0th = basis.eval(t)
basis_1st = basis.eval(t, derivative=1)
basis_2nd = basis.eval(t, derivative=2)

fig.suptitle(f"Etponential basis function and its derivatives for $M={M}$")
axes[0].plot(t, basis_0th, label=r"$\Phi(t)$")
axes[0].legend()
axes[1].plot(t, basis_1st, label=r"$\frac{d\Phi}{dt}(t)$")
axes[1].legend()
axes[2].plot(t, basis_2nd, label=r"$\frac{d^2\Phi}{dt^2}(t)$")
axes[2].legend()
axes[2].set_xlabel(r"$t$")
plt.tight_layout()

plt.show()



"""
This module provides spline basis functions.
"""

import math
import warnings

import numba
import numpy as np


class BasisFunction:
    """
    Base class for all basis functions.

    Parameters
    ----------
    multigenerator : boolean
        This indicates if the basis function generates
        multiple outputs. In practice, this is used to indicate
        if a basis function is meant for a Hermite spline or not.
        Basis functions for Hermite splines return two values
        instead of one.
    support : float
        The support of the function, i.e. the size of the area
        being mapped to non-zero values.
    """

    _unimplemented_message = "This function is not implemented."

    def __init__(self, multigenerator, support):
        self.multigenerator = multigenerator
        self.support = support

    def eval(self, t, derivative=0):
        """
        Evaluate the function at position(s) `t`.

        Parameters
        ----------
        t : float or numpy.array
            The points where the function should be evaluated.
        derivative : [0, 1, 2], default = 0
            Whether to evaluate the function (0) or its first (1)
            or second (2) derivative.

        Returns
        -------
        y : numpy.array
            Values of the function or its first or second derivative
            at position(s) `t`.
        """
        if derivative == 0:
            return self._func(t)
        elif derivative == 1:
            return self._derivative_1(t)
        elif derivative == 2:
            return self._derivative_2(t)
        else:
            raise ValueError(f"derivative has to be 0, 1, or 2 not {derivative}")

    def _func(self, t):
        raise NotImplementedError(BasisFunction._unimplemented_message)

    def _derivative_1(self, t):
        raise NotImplementedError(BasisFunction._unimplemented_message)

    def _derivative_2(self, t):
        raise NotImplementedError(BasisFunction._unimplemented_message)

    def filter_symmetric(self, s):
        """
        The filter used to turn knots into control points for an
        open spline.

        Parameters
        ----------
        s : numpy.array
            An array of knots of shape (n, dim) where n is the number
            of knots and dim is the dimensionality of the codomain, i.e.
            the space in which the curve lives. Note that the knots should
            be padded.

        Returns
        -------
        control_points : numpy.array
            The control points for spline passing through the knots provided.
        """
        raise NotImplementedError(BasisFunction._unimplemented_message)

    def filter_periodic(self, s):
        """
        The filter used to turn knots into control points for a
        closed spline.

        Parameters
        ----------
        s : numpy.array
            An array of knots of shape (n, dim) where n is the number
            of knots and dim is the dimensionality of the codomain, i.e.
            the space in which the curve lives. Note that the knots should
            be padded.

        Returns
        -------
        control_points : numpy.array
            The control points for spline passing through the knots provided.
        """
        raise NotImplementedError(BasisFunction._unimplemented_message)

    def refinement_mask(self):
        """
        This function is needed for local refinement (see [Badoual2016]_).
        Basis splines with the 'local refinement property' can be expressed as a
        linear combination of themselfs. This is useful when you iteratively want
        to refine your spline with additional knots in a given interval.
        This creates a non-uniform spline, which is not supported by splinebox.
        We keep it here incase we ever decide to support non-uniform splines in the
        future.
        """
        raise NotImplementedError(BasisFunction._unimplemented_message)


class B1(BasisFunction):
    r"""
    Basis spline of degree 1.

    .. math::
        f(t) = \begin{cases}1 - \lvert t \rvert & \text{for } -1 \leq t \leq 1 \\ 0 & \text{otherwise}\end{cases}
    """

    def __init__(self):
        super().__init__(False, 2)

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64)], nopython=True, cache=True)
    def _func(t):  # pragma: no cover
        val = 0
        if abs(t) >= 0 and abs(t) < 1:
            val = 1 - abs(t)
        return val

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64)], nopython=True, cache=True)
    def _derivative_1(t):  # pragma: no cover
        val = 0
        if t > -1 and t < 0:
            val = 1
        elif t > 0 and t < 1:
            val = -1
        elif t == 0 or t == -1 or t == 1:
            # This is the gradient you'll get at exactly 0
            val = np.nan
        return val

    @staticmethod
    def _derivative_2(t):
        raise RuntimeError("B1 isn't twice differentiable.")

    @staticmethod
    def filter_symmetric(s):
        return s

    @staticmethod
    def filter_periodic(s):
        return s

    def refinement_mask(self):
        order = int(self.support)
        mask = np.zeros(order + 1)
        _multinomial(order, 2, np.zeros(2, dtype=int), 0, 2, order, mask)
        return mask


class B2(BasisFunction):
    r"""
    Basis spline of degree 2.

    .. math::
        f(t) = \begin{cases}\frac{t^2}{2} + \frac{3}{2} t + \frac{9}{8} & \text{for } -\frac{3}{2} \leq t \leq -\frac{1}{2} \\ -t^2 + \frac{3}{4} & \text{for } -\frac{1}{2} < t \leq \frac{1}{2} \\ \frac{1}{2} t^2 - \frac{3}{2} t + \frac{9}{8} & \text{for } \frac{1}{2} < t \leq \frac{3}{2} \\ 0 & \text{otherwise}\end{cases}
    """

    def __init__(self):
        super().__init__(False, 3)

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64)], nopython=True, cache=True)
    def _func(t):  # pragma: no cover
        val = 0
        if t >= -1.5 and t <= -0.5:
            val = 0.5 * (t**2) + 1.5 * t + 1.125
        elif t > -0.5 and t <= 0.5:
            val = -t * t + 0.75
        elif t > 0.5 and t <= 1.5:
            val = 0.5 * (t**2) - 1.5 * t + 1.125
        return val

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64)], nopython=True, cache=True)
    def _derivative_1(t):  # pragma: no cover
        val = 0
        if t >= -1.5 and t <= -0.5:
            val = t + 1.5
        elif t > -0.5 and t <= 0.5:
            val = -2 * t
        elif t > 0.5 and t <= 1.5:
            val = t - 1.5
        return val

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64)], nopython=True, cache=True)
    def _derivative_2(t):  # pragma: no cover
        val = 0
        if t >= -1.5 and t <= -0.5:
            val = 1
        elif t > -0.5 and t <= 0.5:
            val = -2
        elif t > 0.5 and t <= 1.5:
            val = 1
        return val

    def refinement_mask(self):
        order = int(self.support)
        mask = np.zeros(order + 1)
        _multinomial(order, 2, np.zeros(2, dtype=int), 0, 2, order, mask)
        return mask


class B3(BasisFunction):
    r"""
    Basis spline of degree 3.

    .. math::
        f(t) = \begin{cases}
        \frac{2}{3} - \lvert t \rvert^2 + \frac{\lvert t \rvert^3}{2} & \text{for } 0 \leq \lvert t \rvert < 1 \\
        \frac{1}{6}(2 - \lvert t \rvert)^3 & \text{for } 1 \leq \lvert t \rvert \leq 2 \\
        0 & \text{otherwise}\end{cases}
    """

    def __init__(self):
        super().__init__(False, 4)

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64)], nopython=True, cache=True)
    def _func(t):  # pragma: no cover
        val = 0
        if abs(t) >= 0 and abs(t) < 1:
            val = 2 / 3 - (abs(t) ** 2) + (abs(t) ** 3) / 2
        elif abs(t) >= 1 and abs(t) <= 2:
            val = ((2 - abs(t)) ** 3) / 6
        return val

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64)], nopython=True, cache=True)
    def _derivative_1(t):  # pragma: no cover
        val = 0
        if t >= 0 and t < 1:
            val = -2 * t + 1.5 * t * t
        elif t > -1 and t < 0:
            val = -2 * t - 1.5 * t * t
        elif t >= 1 and t <= 2:
            val = -0.5 * ((2 - t) ** 2)
        elif t >= -2 and t <= -1:
            val = 0.5 * ((2 + t) ** 2)
        return val

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64)], nopython=True, cache=True)
    def _derivative_2(t):  # pragma: no cover
        val = 0
        if t >= 0 and t < 1:
            val = -2 + 3 * t
        elif t > -1 and t < 0:
            val = -2 - 3 * t
        elif t >= 1 and t <= 2:
            val = 2 - t
        elif t >= -2 and t <= -1:
            val = 2 + t
        return val

    @staticmethod
    @numba.jit(nopython=True, nogil=True, cache=True)
    def filter_symmetric(s):  # pragma: no cover
        M = len(s)
        pole = -2 + np.sqrt(3)

        ndim = 1 if s.ndim == 1 else s.shape[1]

        cp = np.zeros((M, ndim))
        eps = 1e-8
        k0 = min(((2 * M) - 2, int(np.ceil(np.log(eps) / np.log(np.abs(pole))))))
        for k in np.arange(k0):
            m = k % (2 * M - 2)
            val = s[2 * M - 2 - m] if m >= M else s[m]
            cp[0] += val * (pole**m)
        cp[0] *= 1 / (1 - (pole ** (2 * M - 2)))

        for k in np.arange(1, M):
            cp[k] = s[k] + pole * cp[k - 1]

        cm = np.zeros((M, ndim))
        cm[M - 1] = cp[M - 1] + (pole * cp[M - 2])
        cm[M - 1] *= pole / ((pole**2) - 1)

        for k in np.arange(M - 2, -1, -1):
            cm[k] = pole * (cm[k + 1] - cp[k])

        c = cm * 6

        shape = c.shape
        c = c.flatten()
        c[np.abs(c) < eps] = 0
        c = c.reshape(shape)
        return c

    @staticmethod
    @numba.jit(nopython=True, nogil=True, cache=True)
    def filter_periodic(s):  # pragma: no cover
        M = len(s)
        pole = -2 + np.sqrt(3)

        ndim = 1 if s.ndim == 1 else s.shape[1]

        cp = np.zeros((M, ndim))
        for k in range(M):
            cp[0] += s[(M - k) % M] * (pole**k)
        cp[0] *= 1 / (1 - (pole**M))

        for k in range(1, M):
            cp[k] = s[k] + pole * cp[k - 1]

        cm = np.zeros((M, ndim))
        for k in range(M):
            cm[M - 1] += (pole**k) * cp[k]
        cm[M - 1] *= pole / (1 - (pole**M))
        cm[M - 1] += cp[M - 1]
        cm[M - 1] *= -pole

        for k in range(M - 2, -1, -1):
            cm[k] = pole * (cm[k + 1] - cp[k])

        c = cm * 6

        eps = 1e-8
        shape = c.shape
        c = c.flatten()
        c[np.abs(c) < eps] = 0
        c = c.reshape(shape)
        return c

    def refinement_mask(self):
        order = int(self.support)
        mask = np.zeros(order + 1)
        _multinomial(order, 2, np.zeros(2, dtype=int), 0, 2, order, mask)
        return mask


class Exponential(BasisFunction):
    r"""
    ???

    .. math::
        f(t) = \begin{cases}
        \frac{L}{\alpha^2} 2 \sin(\frac{\alpha}{2} t)^2 & \text{for } 0 \leq t < 1 \\
        \frac{L}{\alpha^2} (\cos(\alpha (t - 2)) + \cos(\alpha (t - 1)) - 2 \cos(\alpha)) & \text{for } 1 \leq t < 2 \\
        \frac{L}{\alpha^2} 2 \sin(\frac{\alpha}{2} (t - 3))^2 & \text{for } 2 \leq t \leq 3 \\
        0 & \text{otherwise}\end{cases}

    .. math::
        \text{where } L=(\frac{\sin(\pi / M)}{\pi / M})^{-2}\text{ and }t = t + \frac{support}{2}

    .. math::
        f(t) = \begin{cases}
        2 L \sin(\alpha t)^2 & \text{for } 0 \leq t < 1 \\
        L (\cos(2\alpha (t - 2)) + \cos(2\alpha (t - 1)) - 2 \cos(2\alpha)) & \text{for } 1 \leq t < 2 \\
        2 L \sin(\alpha (t - 3))^2 & \text{for } 2 \leq t \leq 3 \\
        0 & \text{otherwise}\end{cases}

    .. math::
        \text{where } \alpha=\frac{\pi}{M}, L=\frac{1}{4 \sin(\alpha)^2} \text{ and }t = t + \frac{3}{2}
    """

    def __init__(self, M):
        super().__init__(False, 3)
        self.M = M

    def _func(self, t):
        return self.__func(t, self.support / 2, self.M)

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64, numba.float64, numba.float64)], nopython=True, cache=True)
    def __func(t, half_support, M):  # pragma: no cover
        t += half_support

        alpha = np.pi / M
        L = 1 / (4 * np.sin(alpha) ** 2)

        val = 0
        if t >= 0 and t < 1:
            val = 2 * np.sin(alpha * t) ** 2
        elif t >= 1 and t < 2:
            val = np.cos(2 * alpha * (t - 2)) + np.cos(2 * alpha * (t - 1)) - 2 * np.cos(2 * alpha)
        elif t >= 2 and t <= 3:
            val = 2 * np.sin(alpha * (t - 3)) ** 2

        return L * val

    def _derivative_1(self, t):
        return self.__derivative_1(t, self.support / 2, self.M)

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64, numba.float64, numba.float64)], nopython=True, cache=True)
    def __derivative_1(t, half_support, M):  # pragma: no cover
        t += half_support

        alpha = np.pi / M
        L = 1 / (4 * np.sin(alpha) ** 2)

        val = 0
        if t >= 0 and t <= 1:
            val = 4 * alpha * np.sin(alpha * t) * np.cos(alpha * t)
        elif t > 1 and t <= 2:
            val = 2 * alpha * (np.sin(2 * alpha * (2 - t)) + np.sin(2 * alpha * (1 - t)))
        elif t > 2 and t <= 3:
            val = 4 * alpha * np.sin(alpha * (t - 3)) * np.cos(alpha * (t - 3))

        return L * val

    def _derivative_2(self, t):
        return self.__derivative_2(t, self.support / 2, self.M)

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64, numba.float64, numba.float64)], nopython=True, cache=True)
    def __derivative_2(t, half_support, M):  # pragma: no cover
        t += half_support

        alpha = np.pi / M
        L = 1 / (4 * np.sin(alpha) ** 2)

        val = 0
        if t >= 0 and t <= 1:
            val = 4 * alpha**2 * (np.cos(alpha * t) ** 2 - np.sin(alpha * t) ** 2)
        elif t > 1 and t <= 2:
            val = -4 * alpha**2 * (np.cos(2 * alpha * (2 - t)) + np.cos(2 * alpha * (1 - t)))
        elif t > 2 and t <= 3:
            val = 4 * alpha**2 * (np.cos(alpha * (t - 3)) ** 2 - np.sin(alpha * (t - 3)) ** 2)

        return L * val

    def filter_symmetric(self, s):
        self.M = len(s)

        b0 = self._func(0)
        b1 = self._func(1)

        return self._filter_symmetric(s, self.M, b0, b1)

    @staticmethod
    @numba.jit(nopython=True, nogil=True, cache=True)
    def _filter_symmetric(s, M, b0, b1):  # pragma: no cover
        ndim = 1 if s.ndim == 1 else s.shape[1]
        pole = (-b0 + np.sqrt(2 * b0 - 1)) / (1 - b0)

        cp = np.zeros((M, ndim))
        eps = 1e-8
        k0 = min(
            (
                (2 * M) - 2,
                int(np.ceil(np.log(eps) / np.log(np.abs(pole)))),
            )
        )
        for k in np.arange(k0):
            m = k % (2 * M - 2)
            val = s[2 * M - 2 - m] if m >= M else s[m]
            cp[0] += val * (pole**m)
        cp[0] *= 1 / (1 - (pole ** (2 * M - 2)))

        for k in range(1, M):
            cp[k] = s[k] + pole * cp[k - 1]

        cm = np.zeros((M, ndim))
        cm[M - 1] = cp[M - 1] + (pole * cp[M - 2])
        cm[M - 1] *= pole / ((pole * pole) - 1)
        for k in range(M - 2, -1, -1):
            cm[k] = pole * (cm[k + 1] - cp[k])

        c = cm / b1

        shape = c.shape
        c = c.flatten()
        c[np.abs(c) < eps] = 0
        c = c.reshape(shape)
        return c

    def filter_periodic(self, s):
        self.M = len(s)
        b0 = self._func(0)
        return self._filter_periodic(s, self.M, b0)

    @staticmethod
    @numba.jit(nopython=True, nogil=True, cache=True)
    def _filter_periodic(s, M, b0):  # pragma: no cover
        ndim = 1 if s.ndim == 1 else s.shape[1]

        pole = (-b0 + np.sqrt(2 * b0 - 1)) / (1 - b0)

        cp = np.zeros((M, ndim))
        cp[0] = s[0]
        for k in range(1, M):
            cp[0] += s[k] * (pole ** (M - k))
        cp[0] *= 1 / (1 - (pole**M))

        for k in range(1, M):
            cp[k] = s[k] + (pole * cp[k - 1])

        cm = np.zeros((M, ndim))
        cm[M - 1] = cp[M - 1]
        for k in range(M - 1):
            cm[M - 1] += cp[k] * (pole ** (k + 1))
        cm[M - 1] *= 1 / (1 - (pole**M))
        cm[M - 1] *= (1 - pole) ** 2

        for k in range(M - 2, -1, -1):
            cm[k] = (pole * cm[k + 1]) + (((1 - pole) ** 2) * cp[k])

        c = cm

        eps = 1e-8
        shape = c.shape
        c = c.flatten()
        c[np.abs(c) < eps] = 0
        c = c.reshape(shape)
        return c

    def refinement_mask(self):
        """
        TODO: This fails the tests. They are set to
        xfail for now but this should be checked at some point.
        """
        warnings.warn(
            "The refinement_mask method of Exponential currently fails our test. This has to be investigated. Double check your results when using it.",
            stacklevel=2,
        )
        order = int(self.support)
        mask = np.zeros(order + 1)

        denominator = 2 ** (order - 1)
        mask[0] = 1 / denominator
        mask[1] = (2 * np.cos(np.pi / self.M) + 1) / denominator
        mask[2] = mask[1]
        mask[3] = 1 / denominator

        return mask


class CatmullRom(BasisFunction):
    """
    ???
    """

    def __init__(self):
        super().__init__(False, 4)

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64)], nopython=True, cache=True)
    def _func(t):  # pragma: no cover
        val = 0
        if np.abs(t) >= 0 and np.abs(t) <= 1:
            val = (3 / 2) * (np.abs(t) ** 3) - (5 / 2) * (np.abs(t) ** 2) + 1
        elif np.abs(t) > 1 and np.abs(t) <= 2:
            val = (-1 / 2) * (np.abs(t) ** 3) + (5 / 2) * (np.abs(t) ** 2) - 4 * np.abs(t) + 2
        return val

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64)], nopython=True, cache=True)
    def _derivative_1(t):  # pragma: no cover
        val = 0
        if t >= 0 and t <= 1:
            val = t * (4.5 * t - 5)
        elif t >= -1 and t < 0:
            val = -t * (4.5 * t + 5)
        elif t > 1 and t <= 2:
            val = -1.5 * t * t + 5 * t - 4
        elif t >= -2 and t < -1:
            val = 1.5 * t * t + 5 * t + 4
        return val

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64)], nopython=True, cache=True)
    def _derivative_2(t):  # pragma: no cover
        val = 0
        if t >= 0 and t <= 1:
            val = 9 * t - 5
        elif t >= -1 and t < 0:
            val = -9 * t - 5
        elif t > 1 and t <= 2:
            val = -3 * t + 5
        elif t >= -2 and t < -1:
            val = 3 * t + 5
        return val

    @staticmethod
    def filter_symmetric(s):
        return s

    @staticmethod
    def filter_periodic(s):
        return s


class CubicHermite(BasisFunction):
    """
    ???
    """

    def __init__(self):
        super().__init__(True, 2)

    def _func(self, t):
        return np.array([self.h31(t), self.h32(t)])

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64)], nopython=True, cache=True)
    def h31(t):  # pragma: no cover
        val = 0
        if t >= 0 and t <= 1:
            val = (1 + (2 * t)) * (t - 1) * (t - 1)
        elif t < 0 and t >= -1:
            val = (1 - (2 * t)) * (t + 1) * (t + 1)
        return val

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64)], nopython=True, cache=True)
    def h32(t):  # pragma: no cover
        val = 0
        if t >= 0 and t <= 1:
            val = t * (t - 1) * (t - 1)
        elif t < 0 and t >= -1:
            val = t * (t + 1) * (t + 1)
        return val

    def _derivative_1(self, t):
        return np.array([self.h31prime(t), self.h32prime(t)])

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64)], nopython=True, cache=True)
    def h31prime(t):  # pragma: no cover
        val = 0
        if t >= 0 and t <= 1:
            val = 6 * t * (t - 1)
        elif t < 0 and t >= -1:
            val = -6 * t * (t + 1)
        return val

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64)], nopython=True, cache=True)
    def h32prime(t):  # pragma: no cover
        val = 0
        if t >= 0 and t <= 1:
            val = 3 * t * t - 4 * t + 1
        elif t < 0 and t >= -1:
            val = 3 * t * t + 4 * t + 1
        return val

    @staticmethod
    def _derivative_2(t):
        raise RuntimeError("CubicHermite isn't twice differentiable.")

    def h31_autocorrelation(self, i, j, M):  # pragma: no cover
        """
        Derived by V. Uhlman during her PhD.
        """
        warnings.warn("This function is untested.", stacklevel=2)
        if self.support > M:
            raise ValueError("Cannot compute h31Autocorrelation for M<" + str(self.support))
            return

        val = 0
        if np.abs(i - j) == 1:
            val = 9 / ((M - 1) * 70)
        elif i == j:
            val = 13 / ((M - 1) * 35) if i == 0 or i == M - 1 else 26 / ((M - 1) * 35)

        return val

    def h31_periodic_autocorrelation(self, n, M):  # pragma: no cover
        """
        Derived by V. Uhlman during her PhD.
        """
        warnings.warn("This function is untested.", stacklevel=2)
        if self.support > M:
            raise ValueError("Cannot compute h31PeriodicAutocorrelation for M<" + str(self.support))
            return

        nmod = np.mod(n, M)
        val = 0
        if nmod == 0:
            val = 26 / (M * 35)
        elif (nmod == 1) or (nmod == M - 1):
            val = 9 / (M * 35) if M == 2 else 9 / (M * 70)

        return val

    def h32_autocorrelation(self, i, j, M):  # pragma: no cover
        """
        Derived by V. Uhlman during her PhD.
        """
        warnings.warn("This function is untested.", stacklevel=2)
        if self.support > M:
            raise ValueError("Cannot compute h32Autocorrelation for M<" + str(self.support))
            return

        val = 0
        if np.abs(i - j) == 1:
            val = -1 / ((M - 1) * 140)
        elif i == j:
            val = 1 / ((M - 1) * 105) if i == 0 or i == M - 1 else 2 / ((M - 1) * 105)

        return val

    def h32_periodic_autocorrelation(self, n, M):  # pragma: no cover
        """
        Derived by V. Uhlman during her PhD.
        """
        warnings.warn("This function is untested.", stacklevel=2)
        if self.support > M:
            raise ValueError("Cannot compute h32PeriodicAutocorrelation for M<" + str(self.support))
            return

        nmod = np.mod(n, M)
        val = 0
        if nmod == 0:
            val = 2 / (M * 105)
        elif (nmod == 1) or (nmod == M - 1):
            val = -1 / (M * 70) if M == 2 else -1 / (M * 140)

        return val

    def h3_crosscorrelation(self, i, j, M):  # pragma: no cover
        """
        Derived by V. Uhlman during her PhD.
        """
        warnings.warn("This function is untested.", stacklevel=2)
        if self.support > M:
            raise ValueError("Cannot compute h3Crosscorrelation for M<" + str(self.support))
            return

        val = 0
        if i - j == 1:
            val = 13 / ((M - 1) * 420)
        elif i - j == -1:
            val = -13 / ((M - 1) * 420)
        elif i == j:
            if i == 0:
                val = 11 / ((M - 1) * 210)
            elif i == M - 1:
                val = -11 / ((M - 1) * 210)

        return val

    def h3_periodic_crosscorrelation(self, n, M):  # pragma: no cover
        """
        Derived by V. Uhlman during her PhD.
        """
        warnings.warn("This function is untested.", stacklevel=2)
        if self.support > M:
            raise ValueError("Cannot compute h3PeriodicCrosscorrelation for M<" + str(self.support))
            return

        nmod = np.mod(n, M)
        val = 0
        if nmod == 1:
            if M != 2:
                val = 13 / (M * 420)
        elif nmod == M - 1:
            val = -13 / (M * 420)

        return val

    @staticmethod
    def filter_symmetric(s):
        return s

    @staticmethod
    def filter_periodic(s):
        return s


class ExponentialHermite(BasisFunction):
    """
    ???
    """

    def __init__(self, M):
        super().__init__(True, 2)
        self.M = M

    def _func(self, t):
        return np.array([self._he31(t, self.M), self._he32(t, self.M)])

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64, numba.float64)], nopython=True, cache=True)
    def _he31(t, M):  # pragma: no cover
        def _g1(t, M):
            val = 0
            if t >= 0 and t <= 1:
                alpha = np.pi / M
                denom = (alpha * np.cos(alpha)) - np.sin(alpha)
                num = (
                    (0.5 * (2 * alpha * np.cos(alpha) - np.sin(alpha)))
                    - (alpha * np.cos(alpha) * t)
                    - (0.5 * np.sin(alpha - (2 * alpha * t)))
                )
                val = num / denom
            return val

        val = _g1(t, M) if t >= 0 else _g1(-t, M)
        return val

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64, numba.float64)], nopython=True, cache=True)
    def _he32(t, M):  # pragma: no cover
        def _g2(t, M):
            val = 0
            if t >= 0 and t <= 1:
                alpha = np.pi / M
                denom = ((alpha * np.cos(alpha)) - np.sin(alpha)) * 8 * alpha * np.sin(alpha)
                num = (
                    -((2 * alpha * np.cos(2 * alpha)) - np.sin(2 * alpha))
                    - (4 * alpha * np.sin(alpha) * np.sin(alpha) * t)
                    - (2 * np.sin(alpha) * np.cos(2 * alpha * (t - 0.5)))
                    + (2 * alpha * np.cos(2 * alpha * (t - 1)))
                )
                val = num / denom
            return val

        val = _g2(t, M) if t >= 0 else -1 * _g2(-t, M)
        return val

    def _derivative_1(self, t):
        return np.array([self._he31prime(t, self.M), self._he32prime(t, self.M)])

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64, numba.float64)], nopython=True, cache=True)
    def _he31prime(t, M):  # pragma: no cover
        def _g1prime(t, M):
            val = 0
            if t >= 0 and t <= 1:
                alpha = np.pi / M
                denom = (alpha * np.cos(alpha)) - np.sin(alpha)
                num = -(alpha * np.cos(alpha)) + (alpha * np.cos(alpha - (2 * alpha * t)))
                val = num / denom
            return val

        val = _g1prime(t, M) if t >= 0 else -1 * _g1prime(-t, M)
        return val

    @staticmethod
    @numba.vectorize([numba.float64(numba.float64, numba.float64)], nopython=True, cache=True)
    def _he32prime(t, M):  # pragma: no cover
        def _g2prime(t, M):
            val = 0
            if t >= 0 and t <= 1:
                alpha = np.pi / M
                denom = ((alpha * np.cos(alpha)) - np.sin(alpha)) * 8 * alpha * np.sin(alpha)
                num = (
                    -(4 * alpha * np.sin(alpha) * np.sin(alpha))
                    + (4 * alpha * np.sin(alpha) * np.sin(2 * alpha * (t - 0.5)))
                    - (4 * alpha**2 * np.sin(2 * alpha * (t - 1)))
                )
                val = num / denom
            return val

        val = _g2prime(t, M) if t >= 0 else _g2prime(-t, M)
        return val

    @staticmethod
    def _derivative_2(x):
        raise RuntimeError("ExponentialHermite isn't twice differentiable.")

    @staticmethod
    def filter_symmetric(s):
        return s

    @staticmethod
    def filter_periodic(s):
        return s


def _multinomial(
    maxValue,
    numberOfCoefficiens,
    kArray,
    iteration,
    dilationFactor,
    order,
    mask,
):
    """
    Recursive function to compute the _multinomial coefficient of (x0+x1+...+xm-1)^N
    This function finds every {k0,...,km-1} such that k0+...+km-1=N
    (cf multinomial theorem on Wikipedia for a detailed explanation)
    """
    if numberOfCoefficiens == 1:
        kArray[iteration] = maxValue

        denominator = 1
        degree = 0
        for k in range(dilationFactor):
            denominator *= math.factorial(kArray[k])
            degree += k * kArray[k]

        coef = math.factorial(order) / denominator
        mask[int(degree)] += coef / (dilationFactor ** (order - 1))

    else:
        for k in range(maxValue + 1):
            kArray[iteration] = k
            _multinomial(
                maxValue - k,
                numberOfCoefficiens - 1,
                kArray,
                iteration + 1,
                dilationFactor,
                order,
                mask,
            )

    return



import math
import unittest.mock

import numpy as np
import pytest
import scipy
import splinebox.basis_functions


def test_check_control_points(non_hermite_spline_curve, coeff_gen):
    spline = non_hermite_spline_curve
    with pytest.raises(RuntimeError):
        spline._check_control_points()
    spline.control_points = coeff_gen(spline.M, spline.basis_function.support, spline.closed)
    spline._check_control_points()


def test_check_control_points_and_tangents(hermite_spline_curve, coeff_gen):
    spline = hermite_spline_curve
    with pytest.raises(RuntimeError):
        spline._check_control_points_and_tangents()
    spline.control_points = coeff_gen(spline.M, spline.basis_function.support, spline.closed)
    with pytest.raises(RuntimeError):
        spline._check_control_points_and_tangents()
    spline.tangents = coeff_gen(spline.M, spline.basis_function.support, spline.closed)
    spline._check_control_points_and_tangents()


def test_minimum_number_of_knots(basis_function):
    M = basis_function.support - 1
    with pytest.raises(RuntimeError):
        splinebox.spline_curves.Spline(M, basis_function)


def test_eval(
    spline_curve, coeff_gen, derivative, eval_positions, is_hermite_spline, not_differentiable_twice, is_interpolating
):
    support = spline_curve.basis_function.support
    half_support = support / 2
    closed = spline_curve.closed

    # Set coefficients and tangents
    spline_curve.control_points = coeff_gen(spline_curve.M, support, closed)
    if is_hermite_spline(spline_curve):
        spline_curve.tangents = coeff_gen(spline_curve.M, support, closed)

    if is_interpolating(spline_curve) and derivative == 0:
        values = spline_curve.eval(np.arange(spline_curve.M), derivative=derivative)
        expected = spline_curve.control_points
        if not closed:
            pad = math.ceil(half_support) - 1
            if pad != 0:
                expected = expected[pad:-pad]
        assert np.allclose(values, expected)

    elif derivative == 2 and not_differentiable_twice(spline_curve):
        with pytest.raises(RuntimeError):
            spline_curve.eval(eval_positions, derivative=derivative)

    else:
        spline_curve.eval(eval_positions, derivative=derivative)

    # Check that the presence of the coefficients and tangents was verified
    if is_hermite_spline(spline_curve):
        spline_curve._check_control_points_and_tangents.assert_called_once()
    else:
        spline_curve._check_control_points.assert_called_once()


def test_set_control_points(spline_curve, is_hermite_spline):
    """
    Test that you can only set coefficients and tangents of the right length.
    """
    if spline_curve.closed:
        expected = spline_curve.M
    else:
        expected = spline_curve.M + 2 * (math.ceil(spline_curve.basis_function.support / 2) - 1)

    for i in range(expected - 2, expected + 2):
        control_points = np.arange(i)
        if len(control_points) != expected:
            with pytest.raises(ValueError):
                spline_curve.control_points = control_points
            if is_hermite_spline(spline_curve):
                with pytest.raises(ValueError):
                    spline_curve.tangents = control_points
        else:
            spline_curve.control_points = control_points
            if is_hermite_spline(spline_curve):
                spline_curve.tangents = control_points


def test_spline_with_hermite_basis(hermite_basis_function, M):
    """
    Test that you cannot use a Hermite basis function for a
    normal spline.
    """
    with pytest.raises(ValueError):
        splinebox.spline_curves.Spline(M, hermite_basis_function)


def test_hermite_spline_with_normal_basis(non_hermite_basis_function, M):
    """
    Test that you can only construct a hermite spline with a hermite basis
    function.
    """
    with pytest.raises(ValueError):
        splinebox.spline_curves.HermiteSpline(M, non_hermite_basis_function)


def test_closed_splines(closed_spline_curve, derivative, coeff_gen, is_hermite_spline, not_differentiable_twice):
    M = closed_spline_curve.M
    support = closed_spline_curve.basis_function.support
    closed_spline_curve.control_points = coeff_gen(M, support, closed=True)
    if is_hermite_spline(closed_spline_curve):
        closed_spline_curve.tangents = coeff_gen(M, support, closed=True)
    if not_differentiable_twice(closed_spline_curve) and derivative == 2:
        return
    assert np.allclose(
        closed_spline_curve.eval(0, derivative=derivative),
        closed_spline_curve.eval(M, derivative=derivative),
        equal_nan=True,
    )


def test_draw():
    spline = splinebox.spline_curves.Spline(M=4, basis_function=splinebox.basis_functions.B1(), closed=True)
    spline._check_control_points = unittest.mock.MagicMock()
    spline.knots = np.array([[1, 1], [1, 2], [2, 2], [2, 1]])

    x = np.linspace(0, 3, 31)
    y = np.linspace(0, 3, 31)

    expected = np.zeros((len(x), len(y)))
    expected[11:20, 11:20] = 1
    expected[10, 10:21] = 0.5
    expected[20, 10:21] = 0.5
    expected[10:21, 10] = 0.5
    expected[10:21, 20] = 0.5

    output = spline.draw(x, y)
    assert np.allclose(output, expected)

    # Check that the presence of the coefficients was verified
    spline._check_control_points.assert_called()

    # You can only draw closed splines
    spline.closed = False
    with pytest.raises(RuntimeError):
        spline.draw(x, y)

    # You can only draw splines in 2D
    spline = splinebox.spline_curves.Spline(M=4, basis_function=splinebox.basis_functions.B1(), closed=True)
    spline.knots = np.array([[1, 1, 0.5], [1, 2, 2], [2, 2, 1], [2, 1, 0]])
    with pytest.raises(RuntimeError):
        spline.draw(x, y)


def test_is_inside():
    # The function itself is already tested via test_draw.
    # Here, we will just test that it raises the correct errors

    # Only makes sense for closed splines.
    spline = splinebox.spline_curves.Spline(M=4, basis_function=splinebox.basis_functions.B1(), closed=False)
    with pytest.raises(RuntimeError):
        spline.is_inside(0, 0)

    # Only makes sense for 2D curves
    spline.closed = True
    spline.knots = np.array([[1, 2, 3], [4, 3, 2], [6, 5, 1], [4, 7, 9]])
    with pytest.raises(RuntimeError):
        spline.is_inside(0, 0)

    # Single value input
    spline.knots = np.array([[1, 1], [1, 2], [2, 2], [2, 1]])
    assert not spline.is_inside(0, 0)
    assert spline.is_inside(1.5, 1.5)

    # x and y vectors with different shapes
    with pytest.raises(ValueError):
        spline.is_inside(np.arange(10), np.arange(15))


def test_arc_length():
    # Create circular spline with radius sqrt(2)
    M = 10
    basis_function = splinebox.basis_functions.Exponential(M)
    spline = splinebox.spline_curves.Spline(M=M, basis_function=basis_function, closed=True)
    spline._check_control_points = unittest.mock.MagicMock()
    phi = np.linspace(0, 2 * np.pi, M + 1)[:-1]
    knots = np.stack([np.cos(phi), np.sin(phi)], axis=-1)
    spline.knots = knots

    ts = np.linspace(0, M, 100)

    # Reset mock in case it was called by any of the methods befor
    spline._check_control_points.reset_mock()

    arc_lengths = [spline.arc_length(t) for t in ts]
    arc_lengths = np.array(arc_lengths)

    # Check that the presence of the coefficients was verified
    spline._check_control_points.assert_called()

    expected = np.linspace(0, 2 * np.pi, 100)
    assert np.allclose(arc_lengths, expected)

    # Check that it is working with a vector for stop
    permutation = np.random.permutation(len(ts))
    arc_lengths = spline.arc_length(ts[permutation])
    assert np.allclose(arc_lengths, expected[permutation])

    # Check that is works with vectors for start and stop
    arc_lengths = spline.arc_length(ts[permutation], np.zeros(len(ts)))
    assert np.allclose(arc_lengths, expected[permutation])


def test_arc_length_to_parameter():
    # Create circular spline with radius sqrt(2)
    M = 100
    basis_function = splinebox.basis_functions.Exponential(M)
    spline = splinebox.spline_curves.Spline(M=M, basis_function=basis_function, closed=True)
    spline._check_control_points = unittest.mock.MagicMock()
    knots = []
    for t in np.linspace(0, 2 * np.pi, M + 1)[:-1]:
        knots.append([np.cos(t), np.sin(t)])
    knots = np.array(knots)
    spline.knots = knots

    ls = np.linspace(0, 2 * np.pi, 15)
    expected = np.linspace(0, M, 15)  # ls / 2 / np.pi * M
    atol = 1e-5

    # Reset mock in case it was called by any of the methods befor
    spline._check_control_points.reset_mock()

    results = spline.arc_length_to_parameter(ls, atol=atol)

    # Check that the presence of the coefficients was verified
    spline._check_control_points.assert_called()

    assert np.allclose(results, expected, atol=1e-3, rtol=0)

    # Create a sawtooth spline
    M = 7
    basis_function = splinebox.basis_functions.B1()
    spline = splinebox.spline_curves.Spline(M=M, basis_function=basis_function, closed=False)
    spline._check_control_points = unittest.mock.MagicMock()
    spline.knots = np.array([[0, 0], [1, 1], [1, 0], [2, 1], [2, 0], [3, 1], [3, 0]])

    rising_length = np.sqrt(2)
    tooth_length = rising_length + 1
    total_length = 3 * tooth_length
    ls = np.linspace(0, total_length, 15)

    def to_param(length):
        n_tooth = length // tooth_length
        residual = length % tooth_length
        partial_tooth_param = residual / rising_length if residual < rising_length else 1 + residual - rising_length
        # there are two knots per tooth
        return n_tooth * 2 + partial_tooth_param

    expected = np.array(list(map(to_param, ls)))
    atol = 1e-4

    # Reset mock in case it was called by any of the methods befor
    spline._check_control_points.reset_mock()

    results = spline.arc_length_to_parameter(ls, atol=atol)

    # Check that the presence of the coefficients was verified
    spline._check_control_points.assert_called()

    assert np.allclose(results, expected, atol=atol)

    # Check that is works for a single value
    expected = to_param(ls[10])
    result = spline.arc_length_to_parameter(ls[10], atol=atol)
    assert np.isclose(result, expected, atol=atol)


def test_translate(initialized_spline_curve, translation_vector):
    spline = initialized_spline_curve
    spline_copy = spline.copy()
    spline_copy._check_control_points = unittest.mock.MagicMock()
    spline_copy.translate(translation_vector)
    # Check that the presence of the coefficients was verified
    spline_copy._check_control_points.assert_called()
    t = np.linspace(0, spline.M, 100) if spline.closed else np.linspace(0, spline.M - 1, 100)
    expected = spline.eval(t) + translation_vector
    assert np.allclose(spline_copy.eval(t), expected)


def test_rotate(initialized_spline_curve, rotation_matrix, is_hermite_spline):
    spline = initialized_spline_curve
    spline_copy = spline.copy()
    spline_copy._check_control_points = unittest.mock.MagicMock()
    if spline.control_points.ndim == 1:
        with pytest.raises(RuntimeError):
            spline_copy.rotate(rotation_matrix)
    else:
        spline_copy.rotate(rotation_matrix, centred=False)
        # Check that the presence of the coefficients was verified
        spline_copy._check_control_points.assert_called()
        t = np.linspace(0, spline.M, 100) if spline.closed else np.linspace(0, spline.M - 1, 100)
        vals = spline.eval(t)
        expected = (rotation_matrix @ vals.T).T
        assert np.allclose(spline_copy.eval(t), expected)

        spline_copy = spline.copy()
        spline_copy._check_control_points = unittest.mock.MagicMock()
        spline_copy.rotate(rotation_matrix, centred=True)
        # Check that the presence of the coefficients was verified
        spline_copy._check_control_points.assert_called()
        t = np.linspace(0, spline.M, 100) if spline.closed else np.linspace(0, spline.M - 1, 100)
        vals = spline.eval(t)
        centring_vector = np.mean(spline.control_points, axis=0)
        vals = vals - centring_vector
        expected = (rotation_matrix @ vals.T).T
        expected = expected + centring_vector
        assert np.allclose(spline_copy.eval(t), expected)


def test_fit(spline_curve, arc_length_parametrization, points, is_hermite_spline):
    # Bool indicating whether spline_curve is a hermite spline or not
    hermite = is_hermite_spline(spline_curve)

    if len(points) < spline_curve.M:
        # The problem is underdetermined
        with pytest.raises(RuntimeError):
            spline_curve.fit(points)
    else:
        spline_curve.fit(points)

        # Keep a copy of the result of fit
        control_points_fit = spline_curve.control_points.copy()
        if hermite:
            tangents_fit = spline_curve.tangents.copy()
            half = len(tangents_fit)

        # Calculate the parameter values for the data points
        if spline_curve.closed:
            t = np.linspace(0, spline_curve.M, len(points) + 1)[:-1]
        else:
            t = np.linspace(0, spline_curve.M - 1, len(points))

        # Add a dimension if the codomain dimensionality is 1
        if points.ndim == 1:
            points = points[:, np.newaxis]

        # Define the objective function for the minimization problem
        def difference_func(x, i):
            if hermite:
                spline_curve.control_points = x[:half]
                spline_curve.tangents = x[half:]
            else:
                spline_curve.control_points = x
            spline_vals = spline_curve.eval(t)
            loss = np.linalg.norm(points[:, i] - spline_vals)
            return loss

        # prepare the initial value for the minimization
        x0 = np.concatenate([control_points_fit, tangents_fit], axis=0) if hermite else control_points_fit

        if x0.ndim == 1:
            x0 = x0[:, np.newaxis]

        # Minimize for each codomain dimension separately
        x = []
        for i in range(x0.shape[1]):
            x.append(scipy.optimize.minimize(difference_func, x0=x0[:, i], args=(i,)).x)

        # Turn minimization result into control_points and tangents for the spline
        x = np.stack(x, axis=-1)
        if hermite:
            control_points_expected = np.squeeze(x[:half])
            tangents_expected = np.squeeze(x[half:])
        else:
            control_points_expected = np.squeeze(x)

        assert np.allclose(control_points_expected, control_points_fit)
        if hermite:
            assert np.allclose(tangents_expected, tangents_fit)


def test_knots(spline_curve, knot_gen, is_hermite_spline, request):
    if isinstance(spline_curve.basis_function, splinebox.basis_functions.B2):
        # The filter_symmetric and filter_periodic are not implemented for B2
        request.node.add_marker(pytest.mark.xfail)

    knots = knot_gen(spline_curve.M)
    pad = math.ceil(spline_curve.basis_function.support / 2) - 1
    if spline_curve.padding_function is None and not spline_curve.closed:
        knots = splinebox.spline_curves.padding_function(knots, pad)
        knots = np.squeeze(knots)
    spline_curve.knots = knots

    if is_hermite_spline(spline_curve):
        # Use finite differences as tangents
        if spline_curve.closed:
            if knots.ndim == 1:
                diff = np.diff(knots, axis=0, append=knots[0])
            else:
                diff = np.diff(knots, axis=0, append=knots[0][np.newaxis, :])
            tangents = (diff + np.roll(diff, 1)) / 2
        else:
            diff = np.diff(knots, axis=0)
            pad = math.ceil(spline_curve.basis_function.support / 2) - 1
            if diff.ndim == 1:
                diff = diff[:, np.newaxis]
            diff = np.pad(diff, ((pad + 1, pad + 1), (0, 0)), mode="constant", constant_values=(0,))
            tangents = (diff[:-1] + diff[1:]) / 2
            tangents = np.squeeze(tangents)
        spline_curve.tangents = tangents

    # This assert makes sense because the knots are not saved in
    # the spline class but are converted to coefficients which are saved.
    if spline_curve.padding_function is None and not spline_curve.closed and pad != 0:
        # We only compare the actual knots because padded knots will not
        # be the same because the of the filtering to find the control points.
        assert np.allclose(knots[pad:-pad], spline_curve.knots[pad:-pad])
    else:
        assert np.allclose(knots, spline_curve.knots)


def test_number_of_knots(spline_curve, knot_gen):
    for n_knots in (spline_curve.M - 1, spline_curve.M + 1):
        knots = knot_gen(n_knots)
        with pytest.raises(ValueError):
            spline_curve.knots = knots

    if not spline_curve.closed and spline_curve.pad > 0:
        knots = knot_gen(spline_curve.M)

        # Test with no padding function
        spline_curve.padding_function = None
        with pytest.raises(ValueError):
            spline_curve.knots = knots

        # Test with custom padding function
        spline_curve.padding_function = lambda knots, pad: knots
        with pytest.raises(ValueError):
            spline_curve.knots = knots


def test_centroid(spline_curve, coeff_gen, is_hermite_spline):
    support = spline_curve.basis_function.support
    closed = spline_curve.closed
    M = spline_curve.M

    control_points = coeff_gen(M, support, closed)
    spline_curve.control_points = control_points
    if is_hermite_spline(spline_curve):
        tangents = coeff_gen(spline_curve.M, support, closed)
        spline_curve.tangents = tangents

    expected = np.mean(control_points, axis=0)
    assert np.allclose(spline_curve._control_points_centroid(), expected)

    spline_curve._check_control_points.assert_called()


def test_scale(initialized_spline_curve, is_hermite_spline):
    spline = initialized_spline_curve
    spline_copy = spline.copy()

    factor = 1.5

    centroid = spline_copy._control_points_centroid()
    spline_copy.translate(-centroid)
    spline_copy.control_points = spline_copy.control_points * factor
    spline_copy.translate(centroid)

    if is_hermite_spline(spline_copy):
        spline_copy.tangents = spline_copy.tangents * factor

    t = np.linspace(0, spline.M, 100) if spline.closed else np.linspace(0, spline.M - 1, 100)

    # Reset mocks in case they were called by another method already
    if is_hermite_spline(spline_copy):
        spline._check_control_points_and_tangents.reset_mock()
    else:
        spline._check_control_points.reset_mock()

    spline.scale(factor)

    # Check that coefficients and tangents were checked
    if is_hermite_spline(spline_copy):
        spline._check_control_points_and_tangents.assert_called()
    else:
        spline._check_control_points.assert_called()

    assert np.allclose(spline.eval(t), spline_copy.eval(t))


def test_curvilinear_reparametrization_energy():
    M = 4
    spline = splinebox.spline_curves.Spline(M, splinebox.B1())

    spline.knots = np.array([[1, 0], [2, 0], [3, 0], [4, 0]])

    # Because of the spacing of the knots the derivative should be constant so we can just compute one value in the middle of the spline.
    derivative_val = spline.eval((spline.M - 1) / 2, derivative=1)

    # To compute the expected value we exploit the constant derivative.
    # We can replace the integration in the definition of the curvilinear reparametrization energy with a multiplication because the derivative is constant.
    length = spline.arc_length()
    c = (length / spline.M) ** 2
    e_curv = (np.linalg.norm(derivative_val) ** 2 - c) ** 2
    expected_val = e_curv * (spline.M - 1) / length**4

    val = spline.curvilinear_reparametrization_energy()

    assert np.isclose(val, expected_val)


def test_curvilinear_reparametrization_energy_translation(initialized_spline_curve, translation_vector):
    """
    Test if the curvilinear reparametrization energy is invariant to translation.
    """
    spline = initialized_spline_curve
    expected = spline.curvilinear_reparametrization_energy()
    spline.translate(translation_vector)
    val = spline.curvilinear_reparametrization_energy()
    assert np.isclose(val, expected)


def test_curvilinear_reparametrization_energy_scale_invariance(initialized_spline_curve):
    """
    The curvilinear reparametrization energy should not change when the entire spline is scaled.
    """
    spline = initialized_spline_curve
    expected = spline.curvilinear_reparametrization_energy()
    spline.scale(0.1)
    val = spline.curvilinear_reparametrization_energy()
    assert np.isclose(val, expected)


def test_curvature():
    # Create a circular spline
    M = 4
    spline = splinebox.spline_curves.Spline(M=M, basis_function=splinebox.basis_functions.Exponential(M), closed=True)
    t = np.linspace(0, M, 100)
    for r in (1, 1.5, 2):
        # 2D
        spline.knots = np.array([[0, r], [r, 0], [0, -r], [-r, 0]])
        curvature = spline.curvature(t)
        assert np.allclose(curvature, 1 / r)

        # If we reverse the direction, the curvature should change sign
        spline.knots = np.array([[0, r], [-r, 0], [0, -r], [r, 0]])
        curvature = spline.curvature(t)
        assert np.allclose(curvature, -1 / r)

        # 3D
        spline.knots = np.array([[0, 0, r], [0, r, 0], [0, 0, -r], [0, -r, 0]])
        curvature = spline.curvature(t)
        assert np.allclose(curvature, 1 / r)

    # 1D
    spline = splinebox.spline_curves.Spline(M=M, basis_function=splinebox.basis_functions.B3(), closed=False)
    spline.knots = np.array([1, 5, 4, 2])
    expected = spline.eval(t, derivative=2) / (1 + spline.eval(t, derivative=1) ** 2) ** (3 / 2)
    result = spline.curvature(t)
    assert np.allclose(result, expected)


def test_normal():
    # Create a circular spline
    M = 4
    spline = splinebox.spline_curves.Spline(M=M, basis_function=splinebox.basis_functions.Exponential(M), closed=True)
    t = np.linspace(0, M, 100)

    theta = np.linspace(0, 2 * np.pi, 100)
    expected = np.stack([np.sin(theta), np.cos(theta)], axis=-1)
    for r in (1, 1.5, 2):
        spline.knots = np.array([[0, r], [r, 0], [0, -r], [-r, 0]])
        normals = spline.normal(t)
        assert np.allclose(normals, expected)

        # Reversing the direction change the direction of the normal vector
        spline.knots = np.array([[0, r], [-r, 0], [0, -r], [r, 0]])
        normals = spline.normal(t[::-1])
        assert np.allclose(normals, -1 * expected)

        # Check that they are normal vectors
        assert np.allclose(np.linalg.norm(normals, axis=1), np.ones(len(t)))



"""
Curvature combs
---------------

In this example we plot the curvature comb of a 2D spline.
Curvature combs are useful to highlight changes in curvature.
"""
import matplotlib.pyplot as plt
import numpy as np
import splinebox

M = 7
spline = splinebox.Spline(M=M, basis_function=splinebox.B3(), closed=False)
spline.knots = np.array([[0, 0], [1, 1], [2.5, 0.5], [3.5, 1.5], [3, 3], [2, 3.5], [1, 2]])

t = np.linspace(0, M - 1, 1000)

vals = spline.eval(t)

curvature = spline.curvature(t)
normals = spline.normal(t)

# You can limit the max height of the curvature comb using the factor d.
max_comb_height = 1
d = max_comb_height / np.max(np.abs(curvature))

comb = vals + d * curvature[:, np.newaxis] * normals

plt.plot(vals[:, 1], vals[:, 0], label="spline")
plt.plot(comb[:, 1], comb[:, 0], label="Curvature comb", color="lightsteelblue")
for p in range(0, len(comb), 7):
    plt.plot([vals[p, 1], comb[p, 1]], [vals[p, 0], comb[p, 0]], color="lightsteelblue")
plt.gca().set_aspect("equal", "box")
plt.legend()
plt.show()



import time

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy
import seaborn as sns
import splinebox

n_repetitions = 10
M = 10
closed = True
ndim = 10
nt = 1000


results = []


for repetition in range(n_repetitions + 1):
    knots = np.random.rand(M, ndim)
    start = time.perf_counter_ns()
    spline = splinebox.Spline(M, splinebox.B3(), closed=closed)
    spline.knots = knots
    stop = time.perf_counter_ns()
    if repetition != 0:
        results.append(
            [
                "splinebox",
                f"{ndim}D",
                M,
                "closed" if closed else "open",
                "Creation",
                stop - start,
            ]
        )

    t = np.linspace(0, M if closed else M - 1, nt)
    start_eval = time.perf_counter_ns()
    spline.eval(t)
    stop_eval = time.perf_counter_ns()
    if repetition != 0:
        results.append(
            [
                "splinebox",
                f"{ndim}D",
                M,
                "closed" if closed else "open",
                "Evaluation",
                stop - start,
            ]
        )

    if closed:
        start = time.perf_counter_ns()
        t_knots = np.arange(M + 1)
        knots_periodic = np.concatenate([knots, knots[0][np.newaxis, :]], axis=0)
        spline = scipy.interpolate.make_interp_spline(t_knots, knots_periodic, k=3, bc_type="periodic")
        stop = time.perf_counter_ns()
    else:
        start = time.perf_counter_ns()
        t_knots = np.arange(M)
        spline = scipy.interpolate.make_interp_spline(t_knots, knots, k=3, bc_type="natural")
        stop = time.perf_counter_ns()

    if repetition != 0:
        results.append(
            [
                "scipy",
                f"{ndim}D",
                M,
                "closed" if closed else "open",
                "Creation",
                stop - start,
            ]
        )

    t = np.linspace(0, M if closed else M - 1, nt)
    start_eval = time.perf_counter_ns()
    spline(t)
    stop_eval = time.perf_counter_ns()
    if repetition != 0:
        results.append(
            [
                "scipy",
                f"{ndim}D",
                M,
                "closed" if closed else "open",
                "Evaluation",
                stop - start,
            ]
        )

    data = np.random.rand(M * 10, ndim)

    start = time.perf_counter_ns()
    spline = splinebox.Spline(M, splinebox.B3(), closed=closed)
    spline.fit(data)
    stop = time.perf_counter_ns()

    if repetition != 0:
        results.append(
            [
                "splinebox",
                f"{ndim}D",
                M,
                "closed" if closed else "open",
                "Fitting",
                stop - start,
            ]
        )

    k = 3
    if closed:
        start = time.perf_counter_ns()
        N = len(data)
        t = np.arange(-k, M + k + 1)
        u = np.linspace(0, M, N + 1)[:-1]
        tck, u = scipy.interpolate.splprep(data.T, k=k, u=u, t=t, task=-1, s=0, per=N)
        stop = time.perf_counter_ns()

    else:
        start = time.perf_counter_ns()
        N = len(data)
        t = np.arange(-k, M + k)
        u = np.linspace(0, M - 1, N)
        tck, u = scipy.interpolate.splprep(data.T, k=k, u=u, t=t, task=-1, s=0)
        stop = time.perf_counter_ns()

    if repetition != 0:
        results.append(
            [
                "scipy",
                f"{ndim}D",
                M,
                "closed" if closed else "open",
                "Fitting",
                stop - start,
            ]
        )


df = pd.DataFrame(results, columns=["Package", "Dimensionality", "Number of knots", "closed", "Task", "time [ns]"])
df["time [ms]"] = df["time [ns]"] / 1000

sns.set_style("darkgrid", {"axes.facecolor": ".9"})
sns.set(font_scale=1.5)
fig, ax = plt.subplots(figsize=(7, 6))
ax.set(yscale="log")
sns.set_palette(sns.color_palette(["#2cb42c", "#0053a6"]))
g = sns.barplot(df, x="Task", y="time [ms]", hue="Package", ax=ax)
g.set(xlabel=None)
plt.tight_layout()
plt.show()



"""
Active contours
===============

This example shows a basic active contours implementation using splinebox.
The goal is to segments the astronaut's head in the example image.
"""

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import scipy
import skimage
import splinebox.basis_functions
import splinebox.spline_curves

# %%
# Let's load the astronaut example image from skimage
img = skimage.data.astronaut()

# %%
# We want our contour to stick to the edges, so we have to
# compute an edge map. To do that we first convert the image
# to gray scale, smooth it to make the edge map less noisy, and
# apply the sobel filter for edge detection.
gray = skimage.color.rgb2gray(img)
smooth = skimage.filters.gaussian(gray, 3, preserve_range=False)
edge = skimage.filters.sobel(smooth)


# %%
# To make calculating the edge energy at non integer locations easier
# we use a surface spline on a regular grid.
# This returns a callable object that we can interogate as follows:
# `edge_energy(x, y)`
edge_energy = scipy.interpolate.RectBivariateSpline(
    np.arange(edge.shape[1]), np.arange(edge.shape[0]), edge, kx=2, ky=2, s=1
)


# %%
# In order to regularize the curvature of our spline we can define an internal energy function
# that scales with the first and second derivative.
def internal_energy(spline, t, alpha, beta):
    return 0.5 * (alpha * spline.eval(t, derivative=1) ** 2 + beta * spline.eval(t, derivative=2) ** 2)


# %%
# Let's initialize our spline with 50 knots that form a circle
# around the astronouts head.
M = 50
s = np.linspace(0, 2 * np.pi, M + 1)[:-1]
y = 100 + 100 * np.sin(s)
x = 220 + 100 * np.cos(s)
knots = np.array([y, x]).T

# %%
# We keep a copy of the initial knots so we can plot them later.
initial_knots = knots.copy()

# %%
# Now, we can construct a B3 spline and adjust its control points (coefficients)
# using the knots we generated above.
spline = splinebox.spline_curves.Spline(M=M, basis_function=splinebox.basis_functions.B3(), closed=True)
print(knots.shape)
spline.knots = knots


# %%
# Here, we set the necessary paramters for active contours:
#
# * :math:`\alpha` controls the contribution of the first derivative to the internal force
# * :math:`\beta` controls the contribution of the second derivative to the internal force
alpha = 0
beta = 0.005

contours = []
external_energies = []


def energy_function(control_points, spline, t, alpha, beta):
    control_points = control_points.reshape((spline.M, -1))
    spline.control_points = control_points
    contour = spline.eval(t)
    contours.append(contour.copy())
    edge_energy_value = np.sum(edge_energy(contour[:, 0], contour[:, 1], grid=False))
    external_energies.append(-edge_energy_value)
    internal_energy_value = np.sum(internal_energy(spline, t, alpha, beta))
    return -edge_energy_value + internal_energy_value


# %%
# The active contours approach consists of iteratively updating our control points (coefficients)
# to minimize the energy.
initial_control_points = spline.control_points.copy()
t = np.linspace(0, M, 400)
result = scipy.optimize.minimize(
    energy_function, initial_control_points.flatten(), method="Powell", args=(spline, t, alpha, beta)
)

# %%
# Inorder to plot the spline as a smooth line, we have to evaluate it
# more densly than just at the knots.
samples = spline.eval(np.linspace(0, len(knots), 400))

final_knots = spline.eval(np.arange(M))

# %%
# Finaly, we can plot the result.
plt.imshow(img)
plt.scatter(initial_knots[:, 1], initial_knots[:, 0], marker="x", color="black", label="initial knots")
contours = contours[::2000]
colors = matplotlib.colormaps["viridis"](np.linspace(0, 1, len(contours)))
for contour, color in zip(contours, colors):
    plt.plot(contour[:, 1], contour[:, 0], color=color, alpha=0.2)
plt.scatter(final_knots[:, 1], final_knots[:, 0], marker="o", color="red", label="final final_knots")
plt.plot(samples[:, 1], samples[:, 0], label="spline", color="red")
plt.legend()
plt.show()



"""
Approximate a noisy signal
--------------------------

Here we add noise to a sinusoidal signal and approximate it.
"""

import matplotlib.pyplot as plt
import numpy as np
import splinebox

# Number of control points/knots
M = 40
# Number of data points
N = 100

x = np.linspace(0, 15, N)
values = np.sin(x) + np.random.rand(N) / 3

basis_function = splinebox.B3()
spline = splinebox.Spline(M, basis_function, closed=False)
spline.fit(values)

ts = np.linspace(0, M - 1, 1000)
spline_y = spline.eval(ts)
spline_x = np.linspace(x.min(), x.max(), len(ts))

plt.scatter(x, values, color="orange")
plt.plot(spline_x, spline_y)
plt.show()



import collections
import copy
import math

import numba
import numpy as np
import scipy.integrate


def padding_function(knots, pad_length):
    """
    This is the default padding function of splinebox.
    It applies constant padding to the ends of the knots.

    Parameters
    ----------
    knots : numpy array
        The knots to be padded.
    pad_length : int
        The amount of padding at each end.

    Returns
    -------
    padded_knots : numpy array
        Array of padded knots.
    """
    # Add constant padding to the ends
    if knots.ndim == 1:
        knots = knots[:, np.newaxis]
    return np.pad(knots, ((pad_length, pad_length), (0, 0)), mode="edge")


class Spline:
    """
    Base class for the construction of a spline.

    Parameters
    ----------
    M : int
        Number of knots.
    basis_function : :class:`splinebox.basis_functions.BasisFunction`
        The basis function used to construct the spline.
    closed : boolean
        Whether or not the spline is closed, i.e. the two ends connect.
    control_points : np.array
        The control points of the spline. Optional, can be provided later.
    padding_function : callable
        A function that accepts an array of knots as the first argument and
        the padding size as the second argument. It should return a padded array.
        If `None`, a padded array has to be supplied when setting the `knots`.
        The default is constant padding with the edge values (see :func:`splinebox.spline_curves.padding_function`).
    """

    _wrong_dimension_msg = "It looks like control_points is a 2D array with second dimension different than two. I don't know how to handle this yet."
    _wrong_array_size_msg = (
        "It looks like control_points is neither a 1 nor a 2D array. I don't know how to handle this yet."
    )
    _no_control_points_msg = "This model doesn't have any coefficients."
    _unimplemented_msg = "This function is not implemented."

    def __init__(self, M, basis_function, closed=False, control_points=None, padding_function=padding_function):
        if basis_function.support <= M:
            self.M = M
        else:
            raise RuntimeError("M must be greater or equal than the spline generator support size.")

        self.basis_function = basis_function
        self.halfSupport = self.basis_function.support / 2
        # Number of additional knots used for padding the ends
        # of an open spline
        self.pad = math.ceil(self.halfSupport) - 1
        self.closed = closed
        self.control_points = control_points
        self.padding_function = padding_function

    def _check_control_points(self):
        """
        Most methods require coefficients to be set before they
        can be used. This helper function checks if the coefficients have been
        set.
        """
        if self.control_points is None:
            raise RuntimeError(self._no_control_points_msg)

    @property
    def control_points(self):
        """
        The control points of this spline, i.e. the c[k]
        in equation :ref:`(1) <theory:eq:1>`.
        """
        return self._control_points

    @control_points.setter
    def control_points(self, values):
        if values is not None:
            n = len(values)
            if self.closed and n != self.M:
                raise ValueError(
                    f"The number of control points must match M for a closed spline. You provided {n} control points for a spline with M={self.M}."
                )
            padded_M = self.M + 2 * self.pad
            if not self.closed and n != padded_M:
                raise ValueError(
                    f"Non-closed splines are padded at the ends, i.e. the effective number of control points is M + 2 * (ceil(support/2) - 1). You provided {n} control points for a spline with M={self.M} and a basis function with support={self.basis_function.support}, expected {padded_M}."
                )
        self._control_points = values

    @property
    def knots(self):
        """
        The knots of this spline, i.e. the values of the spline
        at :math:`t=0,1,...,M`.
        """
        if self.padding_function is None and not self.closed:
            t = np.arange(-self.pad, self.M + self.pad)
        else:
            t = np.arange(self.M)
        return self.eval(t)

    @knots.setter
    def knots(self, values):
        knots = np.array(values)
        n = len(knots)
        if self.closed:
            if n != self.M:
                raise ValueError(
                    f"You provided {n} knots for a closed spline with M={self.M}. Expected {self.M} knots."
                )
            self.control_points = self.basis_function.filter_periodic(knots)
        else:
            padded_M = self.M + 2 * self.pad
            if self.padding_function is None:
                if n != padded_M:
                    raise ValueError(
                        f"If you do not provide a padding function, i.e. `padding_function=None`, you have to pad the knots at the end to have length M + 2 * (ceil(support/2) - 1) before setting them. You provided {n} knots for a spline with M={self.M} and a basis function with support={self.basis_function.support}, expected {padded_M}."
                    )
            else:
                if n != self.M:
                    raise ValueError(f"You provided {n} knots for a spline with M={self.M}. Expected {self.M} knots.")
                knots = self.padding_function(knots, self.pad)
                if len(knots) != padded_M:
                    raise ValueError(
                        f"Incorrect padding. Expected padded to have length {padded_M} instead of {len(knots)}."
                    )

            knots = np.squeeze(knots)

            self.control_points = self.basis_function.filter_symmetric(knots)

    @property
    def basis_function(self):
        """
        The basis function of the spline. Should be an object
        of a specific implementation of the abstract base class
        :class:`splinebox.basis_functions.BasisFunction`.
        """
        return self._basis_function

    @basis_function.setter
    def basis_function(self, value):
        if value.multigenerator:
            raise ValueError(
                "You are trying to construct a Hermite spline using the ordinary `Spline` class. Use the `HermiteSpline` class instead."
            )
        self._basis_function = value

    def copy(self):
        """
        Creats a deep copy of this spline.
        """
        return copy.deepcopy(self)

    def draw(self, x, y):
        """
        Computes whether a point is inside or outside a closed
        spline on a regular grid of points.

        I would ask the user to provide a grid of points directly instead
        of asking for the dimensions. Like that the user can choose how densly
        they want to sample the grid.
        """
        self._check_control_points()

        if self.control_points.ndim != 2 or self.control_points.shape[1] != 2:
            raise RuntimeError("draw() can only be used with 2D curves")

        if not self.closed:
            raise RuntimeError("draw() can only be used with closed curves.")

        xx, yy = np.meshgrid(x, y)
        result = self.is_inside(xx.flatten(), yy.flatten())
        return result.reshape(xx.shape)

    def dtheta(self, t):
        r"""
        Helper function for calculating the winding number.

        `dtheta` is the derivative of the polar coordinate :math:`\theta(t)`

        .. math::
            \theta(t) = arctan \left( \frac{y(t)}{x(t)} \right)

        Differentiation yields:

        .. math::
            \frac{d \theta}{dt} = \frac{1}{r^2} \left( x\frac{dy}{dt} - y\frac{dx}{dt} \right) \text{, where } r^2 = x^2 + y^2
        """
        self._check_control_points()
        r = self.eval(t)
        dr = self.eval(t, derivative=1)
        if r.ndim == 1:
            r = r[np.newaxis, :]
            dr = dr[np.newaxis, :]
        r2 = np.linalg.norm(r, axis=1) ** 2
        if np.any(np.isnan(dr)) or np.isclose(r2, 0):
            # Happens the the spline is not differentiable in this location
            # or when the point is at the origin
            # The solution to return 0 is a bit of a hack but works in practice with
            # the numerical integration
            return np.squeeze(np.zeros(r.shape[0]))
        val = (1.0 / r2) * (r[:, 0] * dr[:, 1] - r[:, 1] * dr[:, 0])
        return np.squeeze(val)

    def is_inside(self, x, y):
        r"""
        Determines if a point with coordinates `x`, `y` is inside the spline.
        Only works for closed 2D curves.

        To determine whether a point is inside or outside the spline, the winding number
        is used:

        .. math::
            wind(\gamma, 0) = \frac{1}{2\pi} \oint_\gamma d\theta = \frac{1}{2\pi} \oint_\gamma \left( \frac{x}{r^2}dy - \frac{y}{r^2}dx \right).

        For a description of :math:`d\theta` check :meth:`splinebox.spline_curves.Spline.dtheta`.

        Parameters
        ----------
        x : numpy.ndarray or float
            x coordinate(s) of point(s)
        y : numpy.ndarray or float
            y coordinate(s) of point(s)

        Returns
        -------
        val : float
            1 if the point is inside, 0.5 if its on the curve and 0 if it is outside the curve.
        """
        if not self.closed:
            raise RuntimeError("isInside() can only be used with closed curves.")
        self._check_control_points()
        if self.control_points.ndim != 2 or self.control_points.shape[1] != 2:
            raise RuntimeError("isInside() can only be used with 2D curves.")

        if isinstance(x, (float, int)):
            x = np.array([x])
        if isinstance(y, (float, int)):
            y = np.array([y])
        if not np.allclose(x.shape, y.shape):
            raise ValueError("x and y need to have the same shape")

        results = np.zeros(x.shape)
        for coord, point in enumerate(np.stack([x, y], axis=-1)):
            spline_copy = self.copy()
            spline_copy.translate(-point)
            winding_number = scipy.integrate.quad(spline_copy.dtheta, 0, spline_copy.M)[0]
            for ref in np.array([[1, 1], [1, 2], [2, 2], [2, 1]]):
                if np.allclose(point, ref):
                    print(ref, winding_number)
            if np.abs(np.abs(winding_number) - 2 * np.pi) < 1e-6:
                results[coord] = 1
            elif np.abs(winding_number) > 1e-6:
                results[coord] = 0.5
            else:
                results[coord] = 0
        results = np.squeeze(results)
        if results.ndim == 0:
            return float(results)
        return results

    def fit(self, points, arc_length_parameterization=False):
        """
        Fit the provided points with the spline using
        least squares.

        Parameters
        ----------
        points : numpy.ndarray
            The data points that should be fit.
        arc_length_parameterization : bool
            Whether or not to space the knots based on the distance
            between the provided points. This is usefull when the
            points are not equally spaced. Default is `False`.
        """
        if len(points) < self.M:
            raise RuntimeError(
                "You provided fewer data points than you spline has knots. For the fit to have a unique solution you need to provide at least as many data points as your spline has knots. Consider adding more data or reducing the number of knots M."
            )
        if arc_length_parameterization:
            raise NotImplementedError
        else:
            t = np.linspace(0, self.M, len(points) + 1)[:-1] if self.closed else np.linspace(0, self.M - 1, len(points))
        tval = self._get_tval(t)
        basis_function_values = self.basis_function.eval(tval, derivative=0)
        self.control_points = np.linalg.lstsq(basis_function_values, points, rcond=None)[0]

    def arc_length(self, stop=None, start=0, epsabs=1e-6, epsrel=1e-6):
        """
        Compute the arc length of the spline between
        the two parameter value specified. If no value for start is give,
        start from the begining of the spline.
        If no value for stop is give, go until the end of the spline.
        When arrays with multiple values are given for start and/or stop,
        an array with all of the arc lengths is returned.

        Parameters
        ----------
        stop : np.array / float (optional)
            Stop point(s) in parameter space.
        start : np.array / float (optional)
            Start point(s) in parameter space.
        epsabs : float (optional)
            Absolute error tolerance. Default is 1e-6.
        epsrel : float (optional)
            Relative error tolerance. Default is 1e-6.
        """
        self._check_control_points()

        if isinstance(stop, collections.abc.Iterable):
            results = np.zeros(len(stop))
            if isinstance(start, collections.abc.Iterable):
                if len(stop) != len(start):
                    raise ValueError(
                        "If you provide array like objects for start and stop, they need to have the same length."
                    )
                for i in range(len(stop)):
                    results[i] = self.arc_length(stop[i], start[i], epsabs=epsabs, epsrel=epsrel)
            else:
                sort_indices = np.argsort(stop)
                sorted_stop = stop[sort_indices]
                for i in range(len(stop)):
                    if i == 0:
                        results[i] = self.arc_length(sorted_stop[i], start, epsabs=epsabs, epsrel=epsrel)
                    else:
                        results[i] = results[i - 1] + self.arc_length(
                            sorted_stop[i], sorted_stop[i - 1], epsabs=epsabs, epsrel=epsrel
                        )
                # Undo the sorting
                results = results[np.argsort(sort_indices)]
            return results

        if stop is None:
            stop = self.M if self.closed else self.M - 1

        if start == stop:
            return 0

        if start > stop:
            start, stop = stop, start

        integral = scipy.integrate.quad(
            lambda t: np.linalg.norm(np.nan_to_num(self.eval(t, derivative=1))),
            start,
            stop,
            epsabs=epsabs,
            epsrel=epsrel,
            maxp1=50,
            limit=100,
        )

        return integral[0]

    def _length_to_parameter_recursion(
        self, s, current_value, lower_bound, upper_bound, intermediate_results=None, atol=1e-4
    ):
        """
        Convert the given arc length s on the curve to a value in parameters space.
        This is done recursively, i.e. check if the point is before or after halfway
        and repeat. It uses binary search.

        Parameters
        ----------
        s : float
            Arc length on the spline.
        current_value : float
            The arc length to the lower bound.
        lower_bound : float
            Lower limit in parameter space.
        upper_bound : float
            Upper limit in parameters space.
        intermediate_results : list
            A list where all computed length parameter pairs are stored.
            This can be used to initialize subsequent conversions more efficiently.
        atol : float
            Absolute precision to which the length is matched.
        """
        self._check_control_points()
        midpoint = lower_bound + (upper_bound - lower_bound) / 2
        midpoint_length = current_value + self.arc_length(lower_bound, midpoint, epsabs=atol)
        if intermediate_results is not None:
            intermediate_results.append((midpoint, midpoint_length))

        if np.isclose(s, midpoint_length, atol=atol, rtol=0):
            return midpoint
        elif s < midpoint_length:
            return self._length_to_parameter_recursion(
                s, current_value, lower_bound, midpoint, intermediate_results, atol
            )
        else:
            return self._length_to_parameter_recursion(
                s, midpoint_length, midpoint, upper_bound, intermediate_results, atol
            )

    def arc_length_to_parameter(self, s, atol=1e-4):
        """
        Convert the arc length `s` to the coresponding value in parameter space.

        Parameters
        ----------
        s : float or np.array
            Length on curve.
        atol : float
            The ablsolute error tolerance.
        """
        self._check_control_points()
        if not isinstance(s, np.ndarray):
            s = np.array([s])
        sort_indices = np.argsort(s)
        results = np.zeros_like(s)

        current_value = 0
        lower_bound = 0
        upper_bound = self.M if self.closed else self.M - 1
        intermediate_results = []

        def upper_bound_key_func(x):
            diff = x[1] - s[i]
            if diff >= 0:
                return diff
            else:
                return np.inf

        for i in sort_indices:
            if len(intermediate_results) > 0:
                upper_bound, upper_bound_len = min(intermediate_results, key=upper_bound_key_func)
                if s[i] > upper_bound_len:
                    upper_bound = self.M if self.closed else self.M - 1

            results[i] = self._length_to_parameter_recursion(
                s[i], current_value, lower_bound, upper_bound, intermediate_results, atol=atol
            )

            lower_bound = results[i]
            _, current_value = min(intermediate_results, key=lambda x: abs(x[0] - lower_bound))

        return np.squeeze(results)

    def sampleArcLength(self, numSamples, dt=False):
        """
        Evaluate the spline equidistantly spaced along its trajectory.
        Perhaps it makes sense to ask the user to provide an array of distances
        instead of the numSamples.
        """
        self._check_control_points()

        if len(self.control_points.shape) == 1 or (
            len(self.control_points.shape) == 2 and self.control_points.shape[1] == 2
        ):
            N = numSamples if self.closed else numSamples - 1
            L = self.arc_length()

            ts = [0]
            for n in range(1, N):
                s = n * L / N
                t = self.lengthToParameter(s)
                ts.append(t)
            if self.closed:
                ts.append(self.M)
            else:
                ts.append(self.M - 1)

            curve = np.array([self.eval(t, dt=dt) for t in ts])

            curve = (
                curve[~np.all(curve == 0)]
                if len(self.control_points.shape) == 1
                else curve[~np.all(curve == 0, axis=1)]
            )

        else:
            raise RuntimeError(self.wrongArraySizeMessage)

        return np.stack(curve)

    def curvilinear_reparametrization_energy(self, epsabs=1e-6, epsrel=1e-6):
        """
        This energy can be used to enforce equal spacing of the knots.

        Implements equation 25 from Jaboc, et al. 2004
        https://infoscience.epfl.ch/record/63117?ln=fr&v=pdf
        In order to make the energy scale invariant,
        we added a factor of length^-4 to the integral.
        """
        arc_length = self.arc_length()
        c = (arc_length / self.M) ** 2
        upper_limit = self.M if self.closed else self.M - 1
        integral = scipy.integrate.quad(
            lambda t: (np.linalg.norm(np.nan_to_num(self.eval(t, derivative=1))) ** 2 - c) ** 2,
            0,
            upper_limit,
            epsabs=epsabs,
            epsrel=epsrel,
            maxp1=50,
            limit=100,
        )
        return integral[0] / arc_length**4

    def curvature(self, t):
        first_deriv = self.eval(t, derivative=1)
        second_deriv = self.eval(t, derivative=2)
        if first_deriv.ndim == 1:
            # This assumes a uniform spline, i.e. the derivative of t is constant.
            first_deriv = np.stack([first_deriv, np.ones(len(t))], axis=-1)
            second_deriv = np.stack([second_deriv, np.zeros(len(t))], axis=-1)
        norm_first_deriv = np.linalg.norm(first_deriv, axis=1)
        norm_second_deriv = np.linalg.norm(second_deriv, axis=1)
        if first_deriv.shape[1] == 2:
            # We use a different formular for the 2D case to get the signed curvature instead of the
            # the unsigned curvature. This is useful when plotting curvature combs.
            nominator = first_deriv[:, 1] * second_deriv[:, 0] - first_deriv[:, 0] * second_deriv[:, 1]
        else:
            nominator = np.sqrt(
                norm_first_deriv**2 * norm_second_deriv**2 - np.sum(first_deriv * second_deriv, axis=1) ** 2
            )
        k = nominator / norm_first_deriv**3
        return np.squeeze(k)

    def normal(self, t):
        """
        Returns the normal vector for 1D and 2D splines.
        The normal vector points to the right of the spline
        when facing in the direction of increasing t.
        """
        self._check_control_points()
        if self.control_points.ndim != 2:
            raise NotImplementedError(
                "The normal vector is only implemented for curves in 2D. Your spline's codomain is 1 dimensional."
            )
        if self.control_points.shape[1] != 2:
            raise RuntimeError(
                f"The normal vector is only defined for curves in 2D. Your spline's codomain is {self.control_points.shape[1]} dimensional."
            )
        first_deriv = self.eval(t, derivative=1)
        normals = (np.array([[0, -1], [1, 0]]) @ first_deriv.T).T
        normals /= np.linalg.norm(normals, axis=1)[:, np.newaxis]
        return normals

    def eval(self, t, derivative=0):
        """
        Evalute the spline or one of its derivatives at
        parameter value(s) `t`.

        Parameters
        ----------
        t : numpy.array, float
            A 1D numpy array or a single float value.
        derivative : int
            Can be 0, 1, 2 for the spline, and its
            first and second derivative respectively.
        """
        self._check_control_points()
        # Get values at which the basis functions have to be evaluated
        tval = self._get_tval(t)
        basis_function_values = self.basis_function.eval(tval, derivative=derivative)
        value = np.matmul(basis_function_values, self.control_points)
        return np.squeeze(value)

    def _get_tval(self, t):
        """
        This is a helper method for `eval`. It is its own method
        to allow :class:`splinebox.spline_curves.HermiteSpline` to
        overwrite the `eval` method using `_get_tval`.
        It is also used in :meth:`splinebox.spline_curves.Spline.fit`
        """
        if not isinstance(t, collections.abc.Iterable):
            t = np.array([t])
        if self.closed:
            # all knot indices
            k = np.arange(self.M)
            # output array for the helper function _wrap_index
            tval = np.full((len(t), len(k)), np.nan)
            # compute the positions at which the basis functions have to be evaluated
            # and save them in tval
            self._wrap_index(t, k, self.halfSupport, self.M, tval)
        else:
            # take into account the padding with additional basis functions
            # for non-closed splines
            k = np.arange(self.M + 2 * self.pad) - self.pad
            k = k[np.newaxis, :]
            t = t[:, np.newaxis]
            # positions at which the basis functions have to be evaluated
            tval = t - k
        return tval

    @staticmethod
    @numba.guvectorize(
        [(numba.float64[:], numba.float64[:], numba.float64, numba.int64, numba.float64[:, :])], "(n),(m),(),()->(n,m)"
    )
    def _wrap_index(ts, ks, halfSupport, M, wrapped_tval):  # pragma: no cover
        """
        Fill the wrapped_tval array whenever a value t
        is affected by the basis function at knot k, taking
        into account that basis functions at the begging/end
        affect positions on the opposite end for closed splines.
        """
        outside_tvalue = halfSupport + 1
        for i, k in enumerate(ks):
            for j, t in enumerate(ts):
                if t >= M + k - halfSupport:
                    # t is close enough to the end to be affected by
                    # the k-th basis function from the beginning
                    wrapped_tval[j, i] = t - M - k
                elif t <= halfSupport - (M - k):
                    # t is close enough to the beginning to be affected
                    # by the k-th basis function, which is close to the end
                    wrapped_tval[j, i] = t + M - k
                elif t > k + halfSupport or t < k - halfSupport:
                    # t is outside the support of the k-th basis function
                    # this can be any value outside the support
                    wrapped_tval[j, i] = outside_tvalue
                else:
                    wrapped_tval[j, i] = t - k

    def _control_points_centroid(self):
        """
        Helper method for :meth:`splinebox.spline_curves.Spline.scale`
        and :meth:`splinebox.spline_curves.Spline.rotate`.
        Computes the centroid of the coefficients.
        """
        self._check_control_points()
        return np.mean(self.control_points, axis=0)

    def translate(self, vector):
        """
        Translates the spline by a `vector`.

        Parameters
        ----------
        vector : numpy.ndarray
            Displacement vector added to the coefficients.
        """
        self._check_control_points()
        self.control_points = self.control_points + vector

    def scale(self, scaling_factor):
        """
        Enlarge or shrink the spline.
        This should probably use :meth:`splinebox.spline_curves.Spline.translate`
        `scalingFactor` can be renamed to `factor`.
        """
        self._check_control_points()
        centroid = self._control_points_centroid()
        self.translate(-centroid)
        self.control_points *= scaling_factor
        self.translate(centroid)

    def rotate(self, rotation_matrix, centred=True):
        """
        Rotate the spline with the provided rotation matrix.

        Parameters
        ----------
        rotation_matrix : numpy.ndarray
            The rotation matrix applied to the spline.
        """
        self._check_control_points()
        if self.control_points.ndim == 1:
            raise RuntimeError("1D splines can not be rotated.")

        if centred:
            centroid = self._control_points_centroid()
            self.translate(-centroid)

        for k in range(len(self.control_points)):
            self.control_points[k] = np.matmul(rotation_matrix, self.control_points[k])

        if centred:
            self.translate(centroid)


class HermiteSpline(Spline):
    """
    Class for the construction of a Hermite spline.

    Parameters
    ----------
    M : int
        Number of control points.
    basis_function : :class:`splinebox.basis_functions.BasisFunction`
        The basis function used to construct the spline.
    closed : boolean
        Whether or not the spline is closed, i.e. the two ends connect.
    """

    _coef_tangent_mismatch_msg = "It looks like control_points and tangents have different shapes."
    _no_tangents_msg = "This spline doesn't have any tangents."

    def __init__(
        self, M, basis_function, closed=False, control_points=None, tangents=None, padding_function=padding_function
    ):
        super().__init__(M, basis_function, closed, control_points=control_points, padding_function=padding_function)
        self.tangents = tangents

    def _check_control_points_and_tangents(self):
        """
        Most methods require coefficients to be set before they
        can be used. This helper function checks if the coefficients have been
        set.
        """
        self._check_control_points()
        if self.tangents is None:
            raise RuntimeError(self._no_tangents_msg)

    @property
    def tangents(self):
        return self._tangents

    @tangents.setter
    def tangents(self, values):
        if values is not None:
            n = len(values)
            if self.closed and n != self.M:
                raise ValueError(
                    f"The number of tangents must match M for a closed spline. You provided {n} tangents for a spline with M={self.M}."
                )
            padded_M = self.M + 2 * self.pad
            if not self.closed and n != padded_M:
                raise ValueError(
                    f"Non-closed splines are padded at the ends, i.e. the effective number of tangents is M + 2 * (ceil(support/2) - 1). You provided {n} tangents for a spline with M={self.M} and a basis function with support={self.basis_function.support}, expected {padded_M}."
                )
        self._tangents = values

    @property
    def basis_function(self):
        return self._basis_function

    @basis_function.setter
    def basis_function(self, value):
        if not value.multigenerator:
            raise ValueError(
                "It looks like you are trying to use a single generator to build a multigenerator spline model."
            )
        self._basis_function = value

    def fit(self, points, arc_length_parameterization=False):
        if len(points) < self.M:
            raise RuntimeError(
                "You provided fewer data points than you spline has knots. For the fit to have a unique solution you need to provide at least as many data points as your spline has knots. Consider adding more data or reducing the number of knots M."
            )
        if arc_length_parameterization:
            raise NotImplementedError
        else:
            t = np.linspace(0, self.M, len(points) + 1)[:-1] if self.closed else np.linspace(0, self.M - 1, len(points))
        tval = self._get_tval(t)
        basis_function_values = self.basis_function.eval(tval, derivative=0)
        basis_function_values = np.concatenate([basis_function_values[0], basis_function_values[1]], axis=1)
        half = self.M if self.closed else self.M + 2 * self.pad
        solution = np.linalg.lstsq(basis_function_values, points, rcond=None)[0]
        self.control_points = solution[:half]
        self.tangents = solution[half:]

    def eval(self, t, derivative=0):
        """
        Evalute the spline or one of its derivatives at
        parameter value(s) `t`.

        Parameters
        ----------
        t : numpy.array, float
            A 1D numpy array or a single float value.
        derivative : int
            Can be 0, 1, 2 for the spline, and its
            first and second derivative respectively.
        """
        self._check_control_points_and_tangents()

        tval = self._get_tval(t)
        basis_function_values = self.basis_function.eval(tval, derivative=derivative)
        value = np.matmul(basis_function_values[0], self.control_points) + np.matmul(
            basis_function_values[1], self.tangents
        )
        return np.squeeze(value)

    def scale(self, scaling_factor):
        self._check_control_points_and_tangents()
        Spline.scale(self, scaling_factor)
        self.tangents *= scaling_factor

    def rotate(self, rotation_matrix, centred=True):
        self._check_control_points_and_tangents()

        Spline.rotate(self, rotation_matrix, centred=centred)

        for k in range(len(self.tangents)):
            self.tangents[k] = np.matmul(rotation_matrix, self.tangents[k])



"""
Generate the splinebox documentation background
-----------------------------------------------

This example shows how the background of the splinebox documentation is made.
"""

import matplotlib.pyplot as plt
import numpy as np
import splinebox
import splinebox.basis_functions
import splinebox.spline_curves

knots = np.array([[0, 1], [0.1, 0.9], [0.2, 0.8], [0.6, 0.7], [0.35, 0.4], [0.8, 0.2], [0.9, 0.1], [1, 0]])
M = len(knots)

spline = splinebox.Spline(M=M, basis_function=splinebox.B3(), closed=False)
spline.knots = knots

ts = np.linspace(0, M - 1, 100)

fig = plt.figure(frameon=False)
fig.set_size_inches(2, 2)
ax = plt.Axes(fig, [0.0, 0.0, 1.0, 1.0])
ax.axis("equal")
ax.set_axis_off()
fig.add_axes(ax)
control_points = np.clip(spline.control_points, 0, 1)
# for the actual background alpha=0.07 was used
alpha = 1
ax.plot(
    control_points[1:-1, 0],
    control_points[1:-1, 1],
    "-o",
    color="forestgreen",
    linewidth=1,
    markersize=1,
    alpha=alpha,
)
vals = spline.eval(ts)
ax.plot(vals[:, 0], vals[:, 1], "-", color="forestgreen", linewidth=1, alpha=alpha)
plt.show()



import matplotlib.pyplot as plt
import numpy as np
import splinebox.basis_functions

fig, axes = plt.subplots(2, 1, sharex=True)

b1 = splinebox.basis_functions.B1()

t = np.linspace(-2, 2, 100)

b1_0th = b1.eval(t)
b1_1st = b1.eval(t, derivative=1)

fig.suptitle("B1 basis function and its derivatives")
axes[0].plot(t, b1_0th, label=r"$\Phi(t)$")
axes[0].legend()
axes[1].plot(t, b1_1st, label=r"$\frac{d\Phi}{dt}(t)$")
axes[1].legend()
axes[1].set_xlabel(r"$t$")
plt.tight_layout()

plt.show()



import matplotlib.pyplot as plt
import numpy as np
import splinebox.basis_functions

fig, axes = plt.subplots(2, 2, sharex=True)

M = 5
basis = splinebox.basis_functions.ExponentialHermite(M=M)

t = np.linspace(-3, 3, 100)

basis_0th = basis.eval(t)
basis_1st = basis.eval(t, derivative=1)

fig.suptitle(f"Exponential Hermite basis function and its derivatives for $M={M}$")
axes[0][0].plot(t, basis_0th[0], label=r"$\Phi_1(t)$")
axes[0][0].legend()
axes[1][0].plot(t, basis_1st[0], label=r"$\frac{d\Phi_1}{dt}(t)$")
axes[1][0].legend()
axes[1][0].set_xlabel(r"$t$")
axes[0][1].plot(t, basis_0th[1], label=r"$\Phi_2(t)$")
axes[0][1].legend()
axes[1][1].plot(t, basis_1st[1], label=r"$\frac{d\Phi_2}{dt}(t)$")
axes[1][1].legend()
axes[1][1].set_xlabel(r"$t$")
plt.tight_layout()

plt.show()



try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"

__all__ = [
    'basis_functions',
    'spline_curves'
]

from .basis_functions import *
from .spline_curves import *



"""
Performance comparison with scipy
---------------------------------

In this example we will compare splinebox's performance
to scipy's splines on the following three tasks:
    - Spline creation
    - Spline evalution
    - Data fitting
"""

import time

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy
import seaborn as sns
import splinebox

# %%
# First we compare the time it takes to create spline
# with a given set of knots and to evalutate it.
# The measurement is repeated 11 and the first repretition
# is discarded, because of the necessary numba compilation.
# The compiled code is cached such that it can be reused
# for future runs.
# Times are computes for different numbers of knots (M),
# open and closed splines, and curves with different dimensionality.

n_repetitions = 10

Ms = np.arange(10, 101, 10)

results_creation = []
results_evaluation = []

for closed in [True, False]:
    for ndim in (1, 2, 3, 10):
        for M in Ms:
            for repetition in range(n_repetitions + 1):
                knots = np.random.rand(M, ndim)

                start = time.perf_counter_ns()
                spline = splinebox.Spline(M, splinebox.B3(), closed=closed)
                spline.knots = knots
                stop = time.perf_counter_ns()
                if repetition != 0:
                    results_creation.append(
                        [
                            "splinebox",
                            f"{ndim}D",
                            M,
                            "closed" if closed else "open",
                            stop - start,
                        ]
                    )

                for nt in [10, 100, 1000, 10000]:
                    t = np.linspace(0, M if closed else M - 1, nt)
                    start_eval = time.perf_counter_ns()
                    spline.eval(t)
                    stop_eval = time.perf_counter_ns()
                    if repetition != 0:
                        results_evaluation.append(
                            [
                                "splinebox",
                                f"{ndim}D",
                                M,
                                nt,
                                "closed" if closed else "open",
                                stop - start,
                            ]
                        )

                if closed:
                    start = time.perf_counter_ns()
                    t_knots = np.arange(M + 1)
                    knots_periodic = np.concatenate([knots, knots[0][np.newaxis, :]], axis=0)
                    spline = scipy.interpolate.make_interp_spline(t_knots, knots_periodic, k=3, bc_type="periodic")
                    stop = time.perf_counter_ns()
                else:
                    start = time.perf_counter_ns()
                    t_knots = np.arange(M)
                    spline = scipy.interpolate.make_interp_spline(t_knots, knots, k=3, bc_type="natural")
                    stop = time.perf_counter_ns()

                if repetition != 0:
                    results_creation.append(
                        [
                            "scipy",
                            f"{ndim}D",
                            M,
                            "closed" if closed else "open",
                            stop - start,
                        ]
                    )

                for nt in [10, 100, 1000, 10000]:
                    t = np.linspace(0, M if closed else M - 1, nt)
                    start_eval = time.perf_counter_ns()
                    spline(t)
                    stop_eval = time.perf_counter_ns()
                    if repetition != 0:
                        results_evaluation.append(
                            [
                                "scipy",
                                f"{ndim}D",
                                M,
                                nt,
                                "closed" if closed else "open",
                                stop - start,
                            ]
                        )

# %%
# Now, that we have collected all of the times, we can turn the results into a data frame and plot the results using seaborn.
df = pd.DataFrame(results_creation, columns=["Package", "Dimensionality", "Number of knots", "closed", "time [ns]"])
for periodicity in ["closed", "open"]:
    fig, ax = plt.subplots()
    ax.set(yscale="log")
    sns.set_palette(sns.color_palette(["#228b18", "#0053a6"]))
    sns.lineplot(
        x="Number of knots",
        y="time [ns]",
        hue="Package",
        style="Dimensionality",
        data=df.loc[df["closed"] == periodicity],
        ax=ax,
    )
    plt.title(f"Creation of {periodicity} splines")
    plt.show()

# %%
# The results show that splinbox outperforms scipy in all condition
# for the task of creating a spline from a given set of knots.

df = pd.DataFrame(
    results_evaluation,
    columns=["Package", "Dimensionality", "Number of knots", "Number of parameter values", "closed", "time [ns]"],
)
for periodicity in ["closed", "open"]:
    fig, ax = plt.subplots(figsize=(7, 7))
    ax.set(yscale="log")
    sns.set_palette(sns.color_palette(["#228b18", "#0053a6"]))
    sns.lineplot(
        x="Number of knots",
        y="time [ns]",
        hue="Package",
        style="Number of parameter values",
        data=df.loc[df["closed"] == periodicity],
        ax=ax,
    )
    plt.title(f"Evaluation of {periodicity} splines")
    plt.show()

# %%
# Once again, splinebox outperforms scipy, in all conditions for the spline evaluation task.
# Lastly, we will compare the fitting (i.e. determining the control points of spline with M knots using least-squares given N > M datapoints) performance.

results_fitting = []
for closed in [True, False]:
    for ndim in (1, 2, 3, 10):
        for M in Ms:
            for repetition in range(n_repetitions + 1):
                data = np.random.rand(M * 10, ndim)

                start = time.perf_counter_ns()
                spline = splinebox.Spline(M, splinebox.B3(), closed=closed)
                spline.fit(data)
                stop = time.perf_counter_ns()

                if repetition != 0:
                    results_fitting.append(
                        [
                            "splinebox",
                            f"{ndim}D",
                            M,
                            "closed" if closed else "open",
                            stop - start,
                        ]
                    )

                k = 3
                if closed:
                    start = time.perf_counter_ns()
                    N = len(data)
                    t = np.arange(-k, M + k + 1)
                    u = np.linspace(0, M, N + 1)[:-1]
                    tck, u = scipy.interpolate.splprep(data.T, k=k, u=u, t=t, task=-1, s=0, per=N)
                    stop = time.perf_counter_ns()

                else:
                    start = time.perf_counter_ns()
                    N = len(data)
                    t = np.arange(-k, M + k)
                    u = np.linspace(0, M - 1, N)
                    tck, u = scipy.interpolate.splprep(data.T, k=k, u=u, t=t, task=-1, s=0)
                    stop = time.perf_counter_ns()

                if repetition != 0:
                    results_fitting.append(
                        [
                            "scipy",
                            f"{ndim}D",
                            M,
                            "closed" if closed else "open",
                            stop - start,
                        ]
                    )
df = pd.DataFrame(
    results_fitting,
    columns=["Package", "Dimensionality", "Number of knots", "closed", "time [ns]"],
)
for periodicity in ["closed", "open"]:
    fig, ax = plt.subplots(figsize=(7, 7))
    ax.set(yscale="log")
    sns.set_palette(sns.color_palette(["#228b18", "#0053a6"]))
    sns.lineplot(
        x="Number of knots",
        y="time [ns]",
        hue="Package",
        style="Dimensionality",
        data=df.loc[df["closed"] == periodicity],
        ax=ax,
    )
    plt.title(f"Fitting of {periodicity} splines")
    plt.show()

# %%
# In the fitting task splinebox is outperformed by scipy's `splprep`,
# but is competetive for splines with relatively few knots (i.e. < 20).



import math

import numpy as np
import pytest
import splinebox.basis_functions


def test_base_class_eval(derivative):
    basis_function = splinebox.basis_functions.BasisFunction(False, 2)
    x = np.arange(10)
    with pytest.raises(NotImplementedError):
        basis_function.eval(x, derivative=derivative)


def test_base_class_filters_and_refinement_mask():
    basis_function = splinebox.basis_functions.BasisFunction(False, 2)
    s = np.arange(10)
    with pytest.raises(NotImplementedError):
        basis_function.filter_symmetric(s)
    with pytest.raises(NotImplementedError):
        basis_function.filter_periodic(s)
    with pytest.raises(NotImplementedError):
        basis_function.refinement_mask()


def test_filters(basis_function, is_interpolating, knot_gen, request):
    if isinstance(basis_function, splinebox.basis_functions.B2):
        # The filter_symmetric and filter_periodic are not implemented for B2
        request.node.add_marker(pytest.mark.xfail)

    s = knot_gen()
    if is_interpolating(basis_function):
        # The filters should not do anything since knot and control points are
        # the same for interpolating splines.
        assert np.allclose(basis_function.filter_symmetric(s), s)
        assert np.allclose(basis_function.filter_periodic(s), s)
    else:
        # Flipping the knot or the control points should give the same result
        assert np.allclose(basis_function.filter_symmetric(s), basis_function.filter_symmetric(s[::-1])[::-1])
        assert np.allclose(basis_function.filter_periodic(s), basis_function.filter_periodic(s[::-1])[::-1])


def test_refinement_mask(basis_function, is_locally_refinable, request):
    if isinstance(basis_function, splinebox.basis_functions.Exponential):
        # The refinement mask of the exponential basis function fails.
        # This might be because the value of M has to be changed after
        # refinement. Since we don't have an imidiate need for the refinement
        # mask, we xfail the tests for now and will revisit this later.
        request.node.add_marker(pytest.mark.xfail)
    if is_locally_refinable(basis_function):
        mask = basis_function.refinement_mask()
        refinement_factor = max(math.floor(len(mask) / 2), 2)
        half_support = basis_function.support / 2
        t = np.linspace(0, basis_function.support, 100) * refinement_factor
        expected = basis_function.eval((t / refinement_factor) - half_support)
        result = np.zeros(len(t))
        # plt.plot(t, expected, label="expected")
        # if isinstance(basis_function, splinebox.basis_functions.Exponential):
        #     basis_function.M *= refinement_factor
        for i in range(len(mask)):
            intermediate = mask[i] * basis_function.eval(t - i - half_support)
            # plt.plot(t, intermediate)
            result += intermediate
        # plt.plot(t, result, label="results")
        # plt.legend()
        # plt.show()
        assert np.allclose(expected, result)
    else:
        with pytest.raises(NotImplementedError):
            basis_function.refinement_mask()


def test_eval_derivative_argument(basis_function, not_differentiable_twice):
    t = np.linspace(-1, 1, 20)
    # Check that 0, 1, 2 don't raise errors
    for derivative in range(3):
        if derivative == 2 and not_differentiable_twice(basis_function):
            continue
        basis_function.eval(t, derivative=derivative)
    with pytest.raises(ValueError):
        basis_function.eval(t, derivative=4)


def test_derivatives(basis_function, derivative, not_differentiable_twice):
    if derivative == 0:
        return

    support = basis_function.support
    x = np.linspace(-support / 2 - 1, support / 2 + 1, 100000)
    y = basis_function.eval(x, derivative=derivative - 1)

    dx = np.diff(x)
    dy = np.diff(y)
    estimated_derivative = dy / dx

    if not_differentiable_twice(basis_function) and derivative == 2:
        # B1, CubicHermite, and ExponentialHermite basis functions are not differentiable twice.
        with pytest.raises(RuntimeError):
            basis_function.eval(x[:-1] + dx / 2, derivative=2)
    else:
        # Check where the estimated derivative is close to the
        # derivative value returned by the method
        close = np.isclose(
            estimated_derivative,
            basis_function.eval(x[:-1] + dx / 2, derivative=derivative),
        )

        # Since the basis functions are not continously differentiable, the
        # estimated_derivative is inaccurate whenever there is a kink in the
        # basis function. These should be isolated values since there are never
        # two kinks right next to each other.
        kernel = np.ones(2)
        if close.ndim == 2:
            # the output of CubicHermite and ExponentialHermite basis functions is 2D
            for i in range(2):
                close[i] = np.convolve(close[i], kernel, mode="same")
        else:
            close = np.convolve(close, kernel, mode="same")

        assert np.all(close > 0)


def test_partition_of_unity(basis_function):
    support = math.ceil(basis_function.support)
    x = np.linspace(0, 1, 10000)
    summed = np.zeros_like(x)
    for k in range(-support, support):
        vals = basis_function.eval(x - k)
        if basis_function.multigenerator:
            vals = vals[0]
        if vals.ndim == 2:
            vals = vals[:, 0]
        summed += vals
    assert np.allclose(summed, np.ones_like(summed))
